#include "aacs_aes.h"

#include <cstring>

#include <openssl/aes.h>
#include <openssl/evp.h>

#include "cmac_aes.h"
#include "cmac.h"


int calculate_subdev_proc_key(const unsigned char *dev_key, /*out*/ unsigned char *subdev_proc_key, unsigned char lmr)  {
   AES_KEY aes_key;

   unsigned char seed[16] = {0x7B,0x10,0x3C,0x5D,0xCB,0x08,0xC4,0xE5,0x1A,0x27,0xB0,0x17,0x99,0x05,0x3B,0xD9};
   seed[15] = seed[15] + lmr;

   AES_set_decrypt_key(dev_key, 128, &aes_key);
   AES_decrypt(seed, subdev_proc_key, &aes_key);

   for (int j = 0; j < 16; j++) {
      subdev_proc_key[j] = subdev_proc_key[j]^seed[j];
   }

   return 1;
}

int decrypt_c_value(const unsigned char *encrypted_c_value, const unsigned char *processing_key, /*out*/ unsigned char *decrypted_c_value, const unsigned char *uv, /*out*/ unsigned char *media_key) {
   AES_KEY aes_key;

   AES_set_decrypt_key(processing_key, 128, &aes_key);
   AES_decrypt(encrypted_c_value, decrypted_c_value, &aes_key);

   for (int j = 0; j < 16; j++) {
      if (j < 12) {
         media_key[j] = decrypted_c_value[j];
      } else {
         media_key[j] = decrypted_c_value[j]^uv[j-12];
      }
   }

   return 1;
}

int decrypt_verification_data(const unsigned char *encrypted_verification_data, const unsigned char *media_key, /*out*/ unsigned char *decrypted_verification_data) {
   AES_KEY aes_key;

   AES_set_decrypt_key(media_key, 128, &aes_key);
   AES_decrypt(encrypted_verification_data, decrypted_verification_data, &aes_key);

   return 1;
}

int calculate_volume_unique_key(const unsigned char *volume_id, const unsigned char *media_key, /*out*/ unsigned char *decrypted_volumeid, /*out*/ unsigned char *volume_unqiue_key) {
   AES_KEY aes_key;

   AES_set_decrypt_key(media_key, 128, &aes_key);
   AES_decrypt(volume_id, decrypted_volumeid, &aes_key);

   for (int j = 0; j < 0 + 16; j++) {
      volume_unqiue_key[j] = volume_id[j]^decrypted_volumeid[j];
   }

   return 1;
}


int decrypt_title_key(const unsigned char *encrypted_title_key, const unsigned char *volume_unqiue_key, const unsigned char *usage_rules, size_t length, /*out*/ unsigned char *decrypted_title_key) {
   AES_KEY aes_key;

   AES_set_decrypt_key(volume_unqiue_key, 128, &aes_key);
   AES_decrypt(encrypted_title_key, decrypted_title_key, &aes_key);

   if (usage_rules != NULL) {
      // For recordables, XOR the AES-H value of the Usage Rules with the decrypted Title Key
      // Calculate AES-H of the Usage Rules
      static const unsigned char h0[16] = { 0x2D, 0xC2, 0xDF, 0x39, 0x42, 0x03, 0x21, 0xD0, 0xCE, 0xF1, 0xFE, 0x23, 0x74, 0x02, 0x9D, 0x95 };
      unsigned char bufin[16];
      unsigned char bufout[16];
      size_t pos = 0;
      memcpy(bufin, h0, 16);
      memset(bufout, '\0', 16);
      for (pos = 0; pos + 16 <= length; pos += 16) {
         AES_set_decrypt_key(usage_rules + pos, 128, &aes_key);
         AES_decrypt(bufin, bufout, &aes_key);
         for (int i = 0; i < 16; i++) {
            bufin[i] = bufout[i] ^ bufin[i];
         }
      }
      // TODO: If the input data is not a multiple of 16 padding must be applied. We skip this step here because the input data *should* always be a multiple of 16!
      // XOR the decrypted Title with the AES-H value
      for (int i = 0; i < 16; i++) {
         decrypted_title_key[i] = decrypted_title_key[i] ^ bufin[i];
      }
   }

   return 1;
}

int calculate_title_key_file_mac(const unsigned char *title_key_file, size_t length, const unsigned char *volume_unqiue_key, /*out*/ unsigned char *title_key_file_mac) {
   AES aes_cipher;
   aes_cipher.makeKey(volume_unqiue_key, 128, DIR_ENCRYPT);

   CMAC cmac(&aes_cipher);
   cmac.init();
   cmac.update(title_key_file, length);
   cmac.finish();

   cmac.getTag(title_key_file_mac, 16);

   return 1;
}


int calculate_title_key_file_hash(const unsigned char *title_key_file, size_t length, /*out*/ unsigned char *title_key_file_hash) {
   EVP_MD_CTX mdctx;
   unsigned int md_len;

   EVP_MD_CTX_init(&mdctx);

   EVP_DigestInit(&mdctx, EVP_sha1());
   EVP_DigestUpdate(&mdctx, title_key_file, length);
   EVP_DigestFinal_ex(&mdctx, title_key_file_hash, &md_len);

   return 1;
}

int calculate_volume_id_mac(const unsigned char *volume_id, size_t length, const unsigned char *bus_key, /*out*/ unsigned char *volume_id_mac) {
   AES aes_cipher;
   aes_cipher.makeKey(bus_key, 128, DIR_ENCRYPT);

   CMAC cmac(&aes_cipher);
   cmac.init();
   cmac.update(volume_id, length);
   cmac.finish();

   cmac.getTag(volume_id_mac, 16);

   return 1;
}
