#undef JNILIB

//
// aacskeys.c : Application to retrieve all aacs keys used to decrypt a disc.
//

// for getcwd and secure template overloads
#if defined(WINDOWS) || defined(_WIN32)
   #ifdef _MSC_VER
      #if _MSC_VER >= 1400
         #ifdef _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES
            #undef _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES
         #endif
         #define _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES 1
         #ifdef _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES_COUNT
            #undef _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES_COUNT
         #endif
         #define _CRT_SECURE_CPP_OVERLOAD_STANDARD_NAMES_COUNT 1
         // Prevent warnings from not overloaded stuff
         // Does only work of overloads are DISABLED!
         //#define _CRT_SECURE_NO_WARNINGS
      #endif
      #include <direct.h>  /* Visual C++ */
   #else
      #include <dir.h>
   #endif
#else
   #include <unistd.h>
   // TODO: _MAX_PATH is the windows version of PATH_MAX and usually 260
   //       It is not advised to use PATH_MAX as buffer size because this can be really big,
   //       what is a good value for that then?
   #define _MAX_PATH 8192
#endif

#include "aacskeys.h"

// General
#include <cstdlib> // for _MAX_PATH, only in Visual Studio?
#include <cstdio>
#include <cstring>

// For variable argument list processing for the printf redirector
#include <cstdarg>

//#include <iostream>
//#include <sstream>

// For directory parsing
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>

// OpenSSL stuff
#include <openssl/bn.h>
#include <openssl/err.h>


#include "aacs_aes.h"
#include "aacs_ecdsa.h"
#include "ioctl.h"
#include "mmc.h"



// No deprecated warnings for insecure CRT functions
#ifdef _MSC_VER
   #pragma warning(disable : 4996)
#endif



char currentpath[_MAX_PATH];

// PS3  2011/08 exposed
unsigned char priv_key[20] = { 0x66, 0x8c, 0x9a, 0x75, 0xee, 0xfc, 0x8d, 0xa4, 0x26, 0x19, 0x38, 0xe2, 0x71, 0x28, 0x50, 0x61, 0xbb, 0x09, 0xf0, 0xdd };
// PS3  2011/08 exposed
unsigned char host_cert[92] = {
   0x02, 0x01, 0x00, 0x5c, 0xff, 0xff, 0x80, 0x00,  0x00, 0x39, 0x00, 0x00, 0x65, 0xea, 0xc9, 0x87,
   0x8b, 0x85, 0xef, 0xf4, 0xd7, 0x7a, 0x62, 0xb1,  0xd6, 0x00, 0x02, 0x4a, 0xce, 0x68, 0xdd, 0x33,
   0x66, 0x88, 0x0e, 0x4f, 0x84, 0x4f, 0x34, 0xb7,  0x7a, 0x05, 0x01, 0x35, 0xa2, 0x0e, 0x73, 0xb6,
   0x26, 0xda, 0xea, 0x51, 0x57, 0xb3, 0x2e, 0xb8,  0x4b, 0xc6, 0xe8, 0x7b, 0x0d, 0xee, 0x4d, 0x83,
   0x3c, 0xea, 0xda, 0x86, 0x12, 0x01, 0x51, 0x00,  0x2c, 0x3c, 0x66, 0xd5, 0x25, 0x6f, 0x71, 0xcf,
   0xa6, 0x8b, 0x7e, 0x55, 0xba, 0x1b, 0x35, 0x1f,  0x34, 0x03, 0x43, 0x4e };

// PS3  2011/08 exposed   valid to P-MKB v28
unsigned char PS3_priv_key[20] = { 0x66, 0x8c, 0x9a, 0x75, 0xee, 0xfc, 0x8d, 0xa4, 0x26, 0x19, 0x38, 0xe2, 0x71, 0x28, 0x50, 0x61, 0xbb, 0x09, 0xf0, 0xdd };
// PS3  2011/08 exposed
unsigned char PS3_host_cert[92] = {
   0x02, 0x01, 0x00, 0x5c, 0xff, 0xff, 0x80, 0x00,  0x00, 0x39, 0x00, 0x00, 0x65, 0xea, 0xc9, 0x87,
   0x8b, 0x85, 0xef, 0xf4, 0xd7, 0x7a, 0x62, 0xb1,  0xd6, 0x00, 0x02, 0x4a, 0xce, 0x68, 0xdd, 0x33,
   0x66, 0x88, 0x0e, 0x4f, 0x84, 0x4f, 0x34, 0xb7,  0x7a, 0x05, 0x01, 0x35, 0xa2, 0x0e, 0x73, 0xb6,
   0x26, 0xda, 0xea, 0x51, 0x57, 0xb3, 0x2e, 0xb8,  0x4b, 0xc6, 0xe8, 0x7b, 0x0d, 0xee, 0x4d, 0x83,
   0x3c, 0xea, 0xda, 0x86, 0x12, 0x01, 0x51, 0x00,  0x2c, 0x3c, 0x66, 0xd5, 0x25, 0x6f, 0x71, 0xcf,
   0xa6, 0x8b, 0x7e, 0x55, 0xba, 0x1b, 0x35, 0x1f,  0x34, 0x03, 0x43, 0x4e };

// 2012/04 exposed   valid to P-MKB v20
unsigned char V20_priv_key[20] = { 0x56, 0x7A, 0x6A, 0x8E, 0xFF, 0xFD, 0x89, 0x67, 0x65, 0x1C, 0xF1, 0xBB, 0x8D, 0x15, 0xED, 0xB6, 0xD2, 0x46, 0x35, 0x55 };
// 2012/04 exposed
unsigned char V20_host_cert[92] = { 
   0x02, 0x00, 0x00, 0x5C, 0xFF, 0xFF, 0x00, 0x00,  0x00, 0x64, 0x00, 0x00, 0x64, 0x40, 0xBE, 0x79,
   0x75, 0x38, 0xE4, 0xFC, 0x36, 0x9F, 0xC5, 0x0B,  0xBE, 0x9F, 0x95, 0xCC, 0x69, 0x43, 0x38, 0x21,
   0x0C, 0xDF, 0xAC, 0xE0, 0xD2, 0xC8, 0x78, 0xBA, 0xB9,  0x6B, 0xB7, 0x2B, 0xA5, 0xA2, 0x9D, 0x0F,
   0x7D, 0x2E, 0x9B, 0x83, 0x6B, 0x4C, 0xE0, 0x67, 0x81,  0xD9, 0x33, 0x54, 0x4E, 0x62, 0x58, 0xF1,
   0xF3, 0x86, 0x68, 0xB4, 0x73, 0x3F, 0x24, 0x63, 0x8C,  0xCB, 0x6F, 0x5B, 0x71, 0x22, 0x0A, 0x22,
   0x20, 0x21, 0x73, 0x67, 0xF8, 0x33, 0x63, 0x5E, 0x97,  0x78, 0x4D, 0x9E };

// KenD00 Original  valid to P-MKB v16
unsigned char ORG_priv_key[20] = { 0x8c, 0x86, 0x47, 0xfe, 0x2a, 0x70, 0xef, 0x03, 0x88, 0xea, 0x9e, 0x43, 0xf4, 0x32, 0xcc, 0x44, 0x1c, 0x6b, 0x10, 0x8c };
// KenD00 Original
unsigned char ORG_host_cert[92] = {
   0x02, 0x00, 0x00, 0x5c, 0xff, 0xff, 0x00, 0x00,  0x00, 0xae, 0x00, 0x00, 0x41, 0x42, 0xa5, 0x41,
   0x1f, 0x1e, 0x63, 0xf1, 0x85, 0x58, 0x1c, 0x87,  0x6b, 0x93, 0x9f, 0xb4, 0x0b, 0x52, 0x3b, 0xf6,
   0x9C, 0x00, 0x4C, 0xA6, 0x9E, 0x04, 0x76, 0x06,  0xEE, 0x51, 0x83, 0xC0, 0xAB, 0xEF, 0x1E, 0x7D,
   0x04, 0xCB, 0x6E, 0x65, 0x26, 0x06, 0x77, 0xE7,  0xB0, 0x57, 0x3D, 0x08, 0xE6, 0x09, 0x57, 0x93,
   0x55, 0x03, 0xED, 0x78, 0xF7, 0xE2, 0x7B, 0x19,  0x0B, 0x4A, 0x7C, 0xAF, 0xCB, 0xAF, 0xF4, 0xA2,
   0x83, 0x64, 0x53, 0xEC, 0xF7, 0x2E, 0x49, 0x66,  0x8D, 0xAF, 0x1D, 0xB9 };

// PowerDVD 7.x  valid for P-MKB v1
unsigned char PD7_priv_key[20] = { 0x47, 0x37, 0x67, 0x60, 0x58, 0xD7, 0x02, 0x94, 0x52, 0x51, 0x4F, 0x0A, 0xB1, 0x86, 0xDC, 0x4C, 0xCA, 0x8C, 0x57, 0x8F };
// PowerDVD 7.x
unsigned char PD7_host_cert[92] = {
   0x02, 0x00, 0x00, 0x5c, 0xff, 0xff, 0x00, 0x00,  0x00, 0x0c, 0x00, 0x00, 0x6e, 0x3d, 0xeb, 0x67,
   0x9b, 0x9a, 0x16, 0xad, 0xfa, 0xa8, 0xe3, 0x08,  0x78, 0x76, 0x7b, 0xa6, 0xeb, 0x2a, 0x9b, 0x41,
   0x53, 0x85, 0xad, 0x11, 0x81, 0xb4, 0x44, 0x6c,  0x31, 0xe9, 0xa5, 0xdd, 0x2a, 0xb8, 0x08, 0xb3,
   0x64, 0xff, 0x15, 0x88, 0x5b, 0xac, 0x49, 0x09,  0x64, 0x31, 0x8c, 0x9b, 0xf8, 0x02, 0x9f, 0xcf,
   0x76, 0xf6, 0x88, 0xa5, 0x4f, 0xbd, 0xa0, 0x3f,  0x6d, 0x93, 0x32, 0xef, 0x04, 0xe5, 0xa6, 0x13,
   0x12, 0xda, 0x85, 0x88, 0x0a, 0x4d, 0x9c, 0xbb,  0x79, 0xd8, 0x60, 0x2e };

unsigned char host_nonce[20] = {0x29, 0x23, 0xbe, 0x84, 0xe1, 0x6c, 0xd6, 0xae, 0x52, 0x90, 0x49, 0xf1, 0xf1, 0xbb, 0xe9, 0xeb, 0xb3, 0xa6, 0xdb, 0x3c};
unsigned char drive_nonce[20];
unsigned char drive_cert[92];
unsigned char drive_key_point[40];
unsigned char drive_key_sig[40];
unsigned char host_key[20] = {0x56,0x13,0xE7,0xF8,0x9B,0x11,0xD9,0xCA,0xA2,0x7B,0x61,0x0A,0x10,0x96,0x33,0x2B,0xEE,0xD8,0x6B,0xC4};
unsigned char host_key_point[40] = {0x8A,0x60,0xC8,0x0B,0xD6,0x0C,0x23,0x60,0x5F,0xBE,0x90,0xB2,0x7B,0xF9,0x6B,0x2D,0xB3,0x81,0x95,0xC1,0x80,0x1F,0x54,0xEB,0x29,0xE0,0xF6,0xEC,0x57,0xAC,0x2B,0x91,0x68,0xE8,0x8B,0x2D,0x56,0x97,0x75,0x08};
unsigned char host_key_sig[40];
unsigned char bus_key[16];
unsigned char mac[16];
unsigned char volume_id_mac[16];
unsigned char title_key_file_hash[20];

const unsigned char *encrypted_c_values = NULL;
const unsigned char *uvs = NULL;
int number_of_uvs = 0;

// Last 32 keys tried
unsigned char possible_keys[32 * 16];

//Device Key
unsigned char device_key[16];
//Possible Processing Keys
unsigned char possible_processing_keys[32 * 16]; // the height of the tree is never higher than 32
//Processing Key
unsigned char processing_key[16]; //= {0x09,0xF9,0x11,0x02,0x9D,0x74,0xE3,0x5B,0xD8,0x41,0x56,0xC5,0x63,0x56,0x88,0xC0};
//Encrypted C Value
unsigned char encrypted_c_value[16]; // = {0x6D,0x02,0xCA,0xC6,0x7B,0x1A,0x7E,0x95,0xC2,0x16,0xEF,0xD4,0xC9,0x28,0x09,0xCF};
//Decrypted C Value
unsigned char decrypted_c_value[16];
// Media Key
unsigned char media_key[16]; // = {0x07,0x4E,0x1F,0xC8,0x8F,0xB9,0xB7,0x80,0xA2,0x25,0xCA,0xA2,0x3B,0xC3,0xDB,0x56};
// uv
unsigned char uv[4]; //= {0x00,0x00,0x00,0x01};
//Encrypted Verification Data (King Kong)
unsigned char encrypted_verification_data[16]; // = {0x87,0xB8,0xA2,0xB7,0xC1,0x0B,0x9F,0xAD,0xF8,0xC4,0x36,0x1E,0x23,0x86,0x59,0xE5};
//Decrypted Verification Data Should Be
unsigned char decrypted_verification_data_should_be[8] = {0x01,0x23,0x45,0x67,0x89,0xAB,0xCD,0xEF};
//Decrypted Verification Data
unsigned char decrypted_verification_data[16];

// LBA of the Protected Area to read from (for recordables only)
unsigned int pa_lba = 0;
// Volume ID
unsigned char volume_id[16]; // = {0x40,0x00,0x09,0x18,0x20,0x06,0x08,0x41,0x00,0x20,0x20,0x20,0x20,0x20,0x00,0x00};
//Decrypted Volume ID
unsigned char decrypted_volumeid[16];
//Volume Unique Key
unsigned char volume_unique_key[16];
unsigned char title_key_file_mac[16];
unsigned char title_key_file_mac_calc[16];

struct title_keys_st {
   unsigned char encrypted_title_key[16];
   unsigned char decrypted_title_key[16];
};
typedef struct title_keys_st TITLE_KEYS;

TITLE_KEYS title_keys[1000];
int nr_of_title_keys = 0;


unsigned char *TitleKeyFileBuffer = NULL;
size_t FileSizeTKF = 0;

unsigned char *MediaKeyFileBuffer = NULL;
size_t FileSizeMKF = 0;


#if defined(WINDOWS) || defined(_WIN32)
const char *pathSeparator = "\\";
#else
const char *pathSeparator = "/";
#endif



// function to return the current working directory
// this is generally the application path
void GetCurrentPath(/*out*/ char *buffer) {
   if (!getcwd(buffer, _MAX_PATH)) {
      strcpy(buffer, ".");
   }
}


// Use a function pointer as printf replacement to be able to redirect the console output
/*
 * This is the target function for console output of the function pointer
 */
int printfc(const char *fmt, ...) {
   int returnVal = 0;
   va_list args;
   va_start(args, fmt);
   returnVal = vprintf(fmt, args);
   va_end(args);
   return returnVal;
}
/*
 * This is the function pointer which should be used instead of the printf function.
 * It points initially to the console output function.
 */
int (*rdprintf)(const char *fmt, ...) = &printfc;


// *** JNI MOD start: global pointers to required java side objects ***
#ifdef JNILIB
// Structure which holds all jni entities which can get used
struct jni_members_st {
   JNIEnv* env;
   jclass aacsException;
   jobject messagePrinter;
   jmethodID print;
   jobject keyData;
   jmethodID getDiscId;
   jmethodID setMek;
   jmethodID setVid;
   jmethodID setVuk;
   jmethodID setTuk;
};
typedef struct jni_members_st JNI_MEMBERS;

// Global pointer to the used JNI_MEMBERS struct
JNI_MEMBERS* globalJni = NULL;

// Buffer to store the sprintf result of the java side output function
// TODO: Which buffer size is a good value?
char sprintfBuffer[4096];
#endif
// *** JNI MOD end ***


void output_key(const unsigned char *key, unsigned int keylength, const char *keyname, unsigned int align, bool hide_private_info) {
   unsigned int i, j = 0;
   size_t namelength = 0;

   (*rdprintf)("%s: ", keyname);
   // +2 because ": " is appended to keyname
   namelength = strlen(keyname) + 2;
   if (align > namelength) {
      for (i = 0; i < align - namelength; i++) {
         (*rdprintf)(" ");
      }
   }
   // Maximum line length is the length of the longest key,
   // full console width doesn't look so good
   for(i = 0; i < keylength; i++)  {
      if (hide_private_info) {
         (*rdprintf)("##");
         if ((i % 20 == 19) && (i != keylength - 1)) {
            (*rdprintf)("\n");
            for (j = 0; j < align; j++) {
               (*rdprintf)(" ");
            }
         }
      } else {
         (*rdprintf)("%02X", key[i]);
         if ((i % 20 == 19) && (i != keylength - 1)) {
            (*rdprintf)("\n");
            for (j = 0; j < align; j++) {
               (*rdprintf)(" ");
            }
         }
      }
   }
   (*rdprintf)("\n");
}


void output_text(const char *text, const char *textname, unsigned int align) {
   unsigned int i = 0;
   const char *text_out = NULL;
   size_t namelength = 0;

   text_out = text;

   (*rdprintf)("%s: ", textname);
   // +2 because ": " is appended to textname
   namelength = strlen(textname) + 2;
   if (align > namelength) {
      for (i = 0; i < align - namelength; i++) {
         (*rdprintf)(" ");
      }
   }
   // Maximum line length is the length of the longest key,
   // full console width doesn't look so good
   //unsigned int width = 80 - align -1;
   unsigned int width = 40;
   while (strlen(text_out) > width) {
      char text_tmp[1000];
      strcpy(text_tmp, text_out);
      text_tmp[width] = 0;
      (*rdprintf)(text_tmp);
      (*rdprintf)("\n");
      for (i = 0; i < align; i++) {
         (*rdprintf)(" ");
      }
      text_out += width;
   }
   (*rdprintf)(text_out);
   (*rdprintf)("\n");
}


const unsigned char *GetRecordFromFile(const unsigned char *FileBuffer, size_t FileSize, unsigned char RecordType, /*out*/ size_t *recordSize = NULL) {
   size_t pos = 0;
   size_t RecordSize = 0;

   while (pos < FileSize) {
      if (pos + 4 <= FileSize) {
         RecordSize = FileBuffer[pos + 1];
         RecordSize <<= 8;
         RecordSize |= FileBuffer[pos + 2];
         RecordSize <<= 8;
         RecordSize |= FileBuffer[pos + 3];

         if (pos + RecordSize <= FileSize) {
            if (FileBuffer[pos] == RecordType) {
               if (recordSize != NULL) {
                  *recordSize = RecordSize;
               }
               return FileBuffer + pos;
            }
         }

         pos += RecordSize;
      } else {
         (*rdprintf)("Error: Broken record at 0x%02X\n", pos);
         return NULL;
      }
   }

   (*rdprintf)("Error: Could not find recordtype 0x%02X\n", RecordType);
   return NULL;
}


void getEncryptedVerificationData(const unsigned char *MediaKeyFileBuffer, size_t FileSize, /*out*/ unsigned char *encrypted_verification_data) {
   const unsigned char *Record = NULL;

   Record = GetRecordFromFile(MediaKeyFileBuffer, FileSize, 0x81);

   if (Record != NULL) {
      memcpy(encrypted_verification_data, Record + 4, 16);
   } else {
      memset(encrypted_verification_data, 0, 16);
   }
}


const unsigned char *getEncryptedCValues(const unsigned char *MediaKeyFileBuffer, size_t FileSize) {
   const unsigned char *Record = NULL;

   // Media Key Data Record
   Record = GetRecordFromFile(MediaKeyFileBuffer, FileSize, 0x05);

   return Record;
}


const unsigned char *getUVs(const unsigned char *MediaKeyFileBuffer, size_t FileSize) {
   const unsigned char *Record = NULL;
   const unsigned char *pos = NULL;
   size_t recordSize = 0;

   // Explicit Subset-Difference Record
   Record = GetRecordFromFile(MediaKeyFileBuffer, FileSize, 0x04, &recordSize);

   //TODO: This is a global variable
   number_of_uvs = 0;

   if (Record != NULL) {
      pos = Record + 4;
      while (pos < MediaKeyFileBuffer + recordSize) {
         //(*rdprintf)("and: %d\n", pos[0] & 0xC0);
         // Check for the end-of-list mark, do not count this entry!
         if ((pos[0] & 0xC0) != 0x00) {
            break;
         }

         pos = pos + 5;
         number_of_uvs++;
      }
      //(*rdprintf)("number_of_uvs: %d\n", number_of_uvs);
   }

   return Record;
}


int LoadMediaKeyBlockFile(const char *drive_name, /*out*/ bool *bluray, /*out*/ bool *recordable, /*out*/ char *aacsdir) {
   FILE *FileHandle = NULL;
   size_t result = 0;
   char filename[_MAX_PATH];
   int ret = 1;

   // TODO: To be on the safer side it would be better to check for BDMV first, but then we
   //       need to check twice for recordables because it might be possible that a MKB_RO.inf
   //       is present on recordables too

   // Check for recordable BDMV
   strcpy(filename, drive_name);
   strcat(filename, pathSeparator);
   strcat(filename, "AACS_mv");
   strcat(filename, pathSeparator);
   strcat(filename, "MKB_RW_mv.inf");
   FileHandle = fopen(filename, "rb");
   if (!FileHandle) {
      // Check for recordable BDAV
      // First check for the MKB_RW.inf
      strcpy(filename, drive_name);
      strcat(filename, pathSeparator);
      strcat(filename, "AACS");
      strcat(filename, pathSeparator);
      strcat(filename, "MKB_RW.inf");
      FileHandle = fopen(filename, "rb");
      // Now check if the Unit Key File is present, MKB_RW.inf alone is also on BD-ROM!
      // This needs to be the second check so that aacsdir gets the right value
      if (FileHandle) {
         strcpy(filename, drive_name);
         strcat(filename, pathSeparator);
         strcat(filename, "AACS");
         strcat(filename, pathSeparator);
         strcat(filename, "AACS_av");
         strcat(filename, pathSeparator);
         strcat(filename, "Unit_Key_RW.inf");
         FILE *FileHandle2 = fopen(filename, "rb");
         if (!FileHandle2) {
            fclose(FileHandle);
            FileHandle = NULL;
         } else {
            fclose(FileHandle2);
         }
      }
   }

   if (!FileHandle) {
      // Check for AVCREC
      strcpy(filename, drive_name);
      strcat(filename, pathSeparator);
      strcat(filename, "AACS_bd");
      strcat(filename, pathSeparator);
      strcat(filename, "MKB_RW.inf");
      FileHandle = fopen(filename, "rb");
      if (FileHandle) {
         strcpy(filename, drive_name);
         strcat(filename, pathSeparator);
         strcat(filename, "AACS_bd");
         strcat(filename, pathSeparator);
         strcat(filename, "AACS_av");
         strcat(filename, pathSeparator);
         strcat(filename, "Unit_Key_RW.inf");
         FILE *FileHandle2 = fopen(filename, "rb");
         if (!FileHandle2) {
            fclose(FileHandle);
            FileHandle = NULL;
         } else {
            fclose(FileHandle2);
         }
      }
   }

   if (!FileHandle) {
      // Check for BDMV
      *recordable = false;
      strcpy(filename, drive_name);
      strcat(filename, pathSeparator);
      strcat(filename, "AACS");
      strcat(filename, pathSeparator);
      strcat(filename, "MKB_RO.inf");
      FileHandle = fopen(filename, "rb");
   } else {
      *recordable = true;
      *bluray = true;
   }

   // Check for HD-DVD
   if (!FileHandle) {
      *bluray = false;
      strcpy(filename, drive_name);
      strcat(filename, pathSeparator);
      strcat(filename, "AACS");
      strcat(filename, pathSeparator);
      strcat(filename, "MKBROM.AACS");
      FileHandle = fopen(filename, "rb");
   } else {
      *bluray = true;
   }

   if (FileHandle == NULL) {
      (*rdprintf)("Error opening Media Key File: %s\n", filename);
      ret = -1;
   } else {
      // Cut off the MKB filename to get the aacs directory alone
      char *cut = strrchr(filename, pathSeparator[0]);
      if (cut != NULL) {
         *cut = '\0';
      }
      strcpy(aacsdir, filename);
      // Get FileSize:
      fseek(FileHandle, 0, SEEK_END);
      FileSizeMKF = ftell(FileHandle);
      rewind(FileHandle);

      // Allocate memory to contain the whole file:
      MediaKeyFileBuffer = (unsigned char *) malloc(FileSizeMKF * sizeof(unsigned char));

      if (MediaKeyFileBuffer == NULL) {
         (*rdprintf)("Error reading Media Key File: could not allocate memory.\n");
         ret = -2;
      } else {
         result = fread(MediaKeyFileBuffer, sizeof(unsigned char), FileSizeMKF, FileHandle);
         if (result != FileSizeMKF) {
            (*rdprintf)("Error reading Media Key File: could not read entire file.\n");
            ret = -3;
         } else {
            getEncryptedVerificationData(MediaKeyFileBuffer, FileSizeMKF, encrypted_verification_data);
         }
      }
      // Close the file
      fclose(FileHandle);
   }

   return ret;
}


void getEncryptedUnitKeys(const unsigned char *TitleKeyFileBuffer, size_t FileSize, /*out*/ TITLE_KEYS *title_keys) {
   size_t pos = 0;
   int i = 0;

   //TODO: Global variable
   nr_of_title_keys = 0;

   if (4 <= FileSize) {
      pos = TitleKeyFileBuffer[0];
      pos <<= 8;
      pos |= TitleKeyFileBuffer[1];
      pos <<= 8;
      pos |= TitleKeyFileBuffer[2];
      pos <<= 8;
      pos |= TitleKeyFileBuffer[3];
      if (pos + 2 <= FileSize) {
         nr_of_title_keys = TitleKeyFileBuffer[pos];
         nr_of_title_keys <<= 8;
         nr_of_title_keys |= TitleKeyFileBuffer[pos + 1];

         //FIXME: this should be done properly...
         if (nr_of_title_keys > 1000) {
            nr_of_title_keys = 1000;
         }

         for (i = 0; i < nr_of_title_keys; i++) {
            size_t keyPos = pos + i * 48 + 48;
            if (keyPos + 16 <= FileSize) {
               memcpy(title_keys[i].encrypted_title_key, TitleKeyFileBuffer + keyPos, 16);
            } else {
               memset(title_keys[i].encrypted_title_key, 0, 16);
            }
         }
      }
   }
}


void getEncryptedTitleKeys(const unsigned char *TitleKeyFileBuffer, size_t FileSize, /*out*/ TITLE_KEYS *title_keys) {
   size_t pos = 128;
   unsigned char bifo = 0;
   int title_key_nr = 0;
   int highest_tk = 0;

   for (title_key_nr = 0; title_key_nr < 64; title_key_nr++) {
      if (pos + 20 <= FileSize) {
         bifo = TitleKeyFileBuffer[pos];
         if ((bifo & 0x80) == 0x80) {
            // Title Key present
            memcpy(title_keys[title_key_nr].encrypted_title_key, TitleKeyFileBuffer + pos + 4, 16);
            // +1 because the limit is excluded
            highest_tk = title_key_nr + 1;
         } else {
            // Title Key not present
            memset(title_keys[title_key_nr].encrypted_title_key, 0, 16);
         }
         pos += 36;
      } else {
         break;
      }
   }
   //TODO: Global variable
   nr_of_title_keys = highest_tk;
}


void getTitleKeyFileMAC(const unsigned char *TitleKeyFileBuffer, size_t FileSize, /*out*/ unsigned char *title_key_file_mac) {
   size_t pos = 2464;

   if (pos + 16 <= FileSize) {
      memcpy(title_key_file_mac, TitleKeyFileBuffer + pos, 16);
   } else {
      memset(title_key_file_mac, 0, 16);
   }
}


int LoadTitleKeyFile(const char *drive_name, bool bluray, bool recordable, bool preferSca, /*out*/ char *tkfpath) {
   FILE *FileHandle = NULL;
   size_t result = 0;
   char filename[_MAX_PATH];
   int ret = 1;

   if (bluray) {
      if (recordable) {
         // Check for recordable BDAV
         strcpy(filename, drive_name);
         strcat(filename, pathSeparator);
         strcat(filename, "AACS");
         strcat(filename, pathSeparator);
         strcat(filename, "AACS_av");
         strcat(filename, pathSeparator);
         strcat(filename, "Unit_Key_RW.inf");
         FileHandle = fopen(filename, "rb");
         if (!FileHandle) {
            // Check for recordable BDMV
            strcpy(filename, drive_name);
            strcat(filename, pathSeparator);
            strcat(filename, "AACS_mv");
            strcat(filename, pathSeparator);
            strcat(filename, "Unit_Key_RW.inf");
            FileHandle = fopen(filename, "rb");
            if (!FileHandle) {
               // Check for AVCREC
               strcpy(filename, drive_name);
               strcat(filename, pathSeparator);
               strcat(filename, "AACS_bd");
               strcat(filename, pathSeparator);
               strcat(filename, "AACS_av");
               strcat(filename, pathSeparator);
               strcat(filename, "Unit_Key_RW.inf");
               FileHandle = fopen(filename, "rb");
            }
         }
      } else {
         // Check for BDMV
         strcpy(filename, drive_name);
         strcat(filename, pathSeparator);
         strcat(filename, "AACS");
         strcat(filename, pathSeparator);
         strcat(filename, "Unit_Key_RO.inf");
         FileHandle = fopen(filename, "rb");
      }
   } else {
      // Check for Standard Content
      strcpy(filename, drive_name);
      strcat(filename, pathSeparator);
      strcat(filename, "AACS");
      strcat(filename, pathSeparator);
      strcat(filename, "VTKF.AACS");
      FileHandle = fopen(filename, "rb");
      if (!FileHandle || !preferSca) {
         // Check for Advanced Content
         strcpy(filename, drive_name);
         strcat(filename, pathSeparator);
         strcat(filename, "AACS");
         DIR *dir = opendir(filename);
         if (dir != NULL) {
            struct dirent *entry;
            char tkffile[13];
            strcpy(tkffile, "VTKF999.AACS");
            while ((entry = readdir(dir)) != NULL) {
               if (strlen(entry->d_name) == 12) {
                  if (!strncmp(entry->d_name, "VTKF", 4) && isdigit(entry->d_name[4]) && isdigit(entry->d_name[5]) && isdigit(entry->d_name[6]) && !strcmp(entry->d_name + 7, ".AACS")) {
                     if (strcmp(entry->d_name, tkffile) < 0) {
                        strcpy(tkffile, entry->d_name);
                     }
                  }
               }
            }
            closedir(dir);
            strcat(filename, pathSeparator);
            strcat(filename, tkffile);
            //(*rdprintf)("Using TKF: %s\n", filename);
         } else if (errno == ENOENT) {
            return -4;
         } else if (errno == EINVAL) {
            return -5;
         } else {
            return -6;
         }

         FILE* testHandle = fopen(filename, "rb");
         if (FileHandle) {
            if (testHandle) {
               fclose(FileHandle);
               FileHandle = testHandle;
            }
         } else {
            FileHandle = testHandle;
         }
      }
   }

   if (FileHandle == NULL) {
      (*rdprintf)("Error opening Title Key / Unit Key File: %s\n", filename);
      ret = -1;
   } else {
      // Save the path to the Title / Unit Key file, this is needed for recordables
      // TODO: Maybe the drive path needs to be cut
      strcpy(tkfpath, filename);
      // Get FileSize:
      fseek(FileHandle, 0, SEEK_END);
      FileSizeTKF = ftell (FileHandle);
      rewind(FileHandle);

      // Allocate memory to contain the whole file:
      TitleKeyFileBuffer = (unsigned char *) malloc(FileSizeTKF * sizeof(unsigned char));
      if (TitleKeyFileBuffer == NULL)  {
         (*rdprintf)("Error reading Title Key File: could not allocate memory.\n");
         ret = -2;
      } else {
         result = fread(TitleKeyFileBuffer, sizeof(unsigned char), FileSizeTKF, FileHandle);
         if (result != FileSizeTKF) {
            (*rdprintf)("Error reading Title Key File: could not read entire file.\n");
            ret = -3;
         } else {
            if (bluray) {
               getEncryptedUnitKeys(TitleKeyFileBuffer, FileSizeTKF, title_keys);
            } else {
               getEncryptedTitleKeys(TitleKeyFileBuffer, FileSizeTKF, title_keys);
               getTitleKeyFileMAC(TitleKeyFileBuffer, FileSizeTKF, title_key_file_mac);
            }
         }
      }
      // Close the file
      fclose(FileHandle);
   }

   return ret;
}


// This is essentially an implementation of the official algo described at 3.2.4 (apart from checking whether a subset-diff applies to a device, since we have no devicenr and we simply assume this to be the case)
void calculate_processing_key(const unsigned char *device_key, /*out*/ unsigned char *processing_key, long uv_long, long v_mask, long u_mask, long device_key_uv_long, long device_key_v_mask, long device_key_u_mask)
{

   // This is from 3.2.4: (mu == m'u) and ((uv & m'v) == (uv' & m'v))
   //if ((u_mask == device_key_u_mask) && ((uv_long & device_key_v_mask) == (device_key_uv_long & device_key_v_mask)))
   if ((u_mask == device_key_u_mask) && (v_mask == device_key_v_mask))
   {
      // This is the (sub) device just before the processing key so calculate the proc key
      calculate_subdev_proc_key(device_key, processing_key, 1);
   }
   else
   {
      // We have to "move" southwards until we reach the position of the desired processing key

      // 1. Initialization. m = stored v mask m'v, D = starting Device Key

      long mask_m;
      unsigned char device_key_D[16];
      unsigned char device_key_D_tmp[16];

      mask_m = device_key_v_mask;
      memcpy (device_key_D, device_key, 16);

      int notready = 1;
      //int cnt = 0;
      while (notready)
      {
         // 2. Use AES-G3 on D, as described above, to determine a left subsidiary Device Key, a Processing Key, and a right subsidiary Device Key.

         // --> we skip this here and do it later (better performance)

         // 3. Look at the most significant zero bit in m. If the corresponding bit in the incoming uv number is 0, D = left
         // subsidiary Device Key from step 2. Otherwise, D = right subsidiary Device Key from step 2

         long mask_right = 1;
         // The loop will always be executed at least once, but initalize this to avoid a warning
         long mask_check = 0;
         while ((mask_m | mask_right) != -1L)  // check if all bits are 1, this way works for 32 and 64 bit
         {
            mask_check = ~(mask_m | mask_right);
            mask_right <<= 1; // move 1 bits to left
            mask_right++; // make sure the lowest bit is set
         }

         //(*rdprintf) ("mask_m %08X\n", mask_check);
         // mask_check now contains 1 bit which is the most significant zero bit in m.
         // Now check if this bit is 1 or 0 (in the incoming uv) to determine left or right sub key to be

         if ((mask_check & uv_long) == 0)
         {
            // Calculate left subkey
            calculate_subdev_proc_key(device_key_D, device_key_D_tmp, 0);
            //(*rdprintf) ("left %d\n", cnt++);
         }
         else
         {
            // Calculate right subkey
            calculate_subdev_proc_key(device_key_D, device_key_D_tmp, 2);
            //(*rdprintf) ("right %d\n", cnt++);
         }

         // copy the new sub dev key to D
         memcpy (device_key_D, device_key_D_tmp, 16);

         // 4. Iteration. Arithmetic shift m right one bit. If it does not equal the incoming v mask mv, repeat starting at step 2.

         mask_m >>= 1;
         mask_m |= 1L << ((sizeof(mask_m) * 8) - 1); // Ensure the highest bit is set, this way works under 32 and 64 bit


         if (mask_m == v_mask)
         {
            notready = 0;
         }
      }

      // Finally calculate the processing key (using the last sub device key)
      calculate_subdev_proc_key(device_key_D, processing_key, 1);

   }

}


int get_possible_processing_keys(const unsigned char *device_key, unsigned char umask_nr) {
   int nr_of_possible_processing_keys = 0;
   //int cnt = 0;

   long v_mask = -1L; // to set all bits 1 under 32 and 64 bit depth!
   long u_mask = -1L; // to set all bits 1 under 32 and 64 bit depth!
   long uv_long;
   long device_key_v_mask;
   long device_key_u_mask;
   long device_key_uv_long;
   uv_long = uv[0];
   uv_long <<= 8;
   uv_long |= uv[1];
   uv_long <<= 8;
   uv_long |= uv[2];
   uv_long <<= 8;
   uv_long |= uv[3];
   while ((uv_long & ~v_mask) == 0) v_mask <<= 1;

   u_mask <<= umask_nr;

   // start with the assumption that this key is the (sub) device key just above the processing key
   device_key_v_mask = v_mask;
   device_key_u_mask = u_mask;
   device_key_uv_long = uv_long;

   while (device_key_v_mask != u_mask) // go on until the the top of this tree is reached
   {
      /*if (cnt <= 50)
      {
      (*rdprintf) ("\n%d ", cnt);
      (*rdprintf) ("uv %08X ", uv_long);
      (*rdprintf) ("v_mask %08X ", v_mask);
      (*rdprintf) ("u_mask %08X\n", u_mask);
      (*rdprintf) ("d_uv %08X ", device_key_uv_long);
      (*rdprintf) ("d_v_mask %08X ", device_key_v_mask);
      (*rdprintf) ("d_u_mask %08X\n", device_key_u_mask);
      cnt++;
      }*/

      //(*rdprintf) ("%d calculate_processing_key\n", cnt++);
      calculate_processing_key(device_key, processing_key, uv_long, v_mask, u_mask, device_key_uv_long, device_key_v_mask, device_key_u_mask);

      memcpy(possible_processing_keys + 16 * nr_of_possible_processing_keys, processing_key, 16);
      nr_of_possible_processing_keys++;

      device_key_v_mask <<= 1; // assume the key is located a little higher up in the tree
   }

   return nr_of_possible_processing_keys;
}


int check_processing_key() {
   int found = 0;
   // FIXME: These calculations crash, why? Wrong openssl?
   // Decrypt the c value into the media key (using the processing key and the corresponding uv)
   decrypt_c_value(encrypted_c_value, processing_key, decrypted_c_value, uv, media_key);

   // Decrypt the verification data (using the media key)
   decrypt_verification_data(encrypted_verification_data, media_key, decrypted_verification_data);

   // Check if all is correct
   if (!memcmp(decrypted_verification_data_should_be, decrypted_verification_data, 8))
   {
      found = 1;
   }

   return found;
}


// Don't use the return value as bool because negative values can get returned which evaluate as TRUE!!
int process_media_key_block (int align, bool verbose) {
   int found = 0;

   FILE *FileHandle = NULL;
   char filename[_MAX_PATH];
   BIGNUM *key_bn = NULL;

   //TODO: Global variables
   encrypted_c_values = getEncryptedCValues(MediaKeyFileBuffer, FileSizeMKF);
   uvs = getUVs(MediaKeyFileBuffer, FileSizeMKF);

   if ((encrypted_c_values == NULL) || (uvs == NULL)) {
      (*rdprintf)("Broken MKB!\n");
      found = -2;
   } else {
      strcpy(filename, currentpath);
      strcat(filename, pathSeparator);
      strcat(filename, "ProcessingDeviceKeysSimple.txt");
      FileHandle = fopen(filename, "r");
      if (!FileHandle) {
         (*rdprintf)("Could not open file: %s\n", filename);
         found = -1;
      } else {
         /*if (verbose) {
         output_key(encrypted_c_values + 4, 16, "First C-value", align, 0);
         output_key(uvs + 4, 1, "First u mask nr", align, 0);
         output_key(uvs + 4 + 1, 4, "First uv", align, 0);
         }*/

         const size_t linebuffer_size = 1024;
         char line[linebuffer_size];
         unsigned char key_bin[16];
         int keys_tried = 0;
         //int PK_cnt = 0;
         while (!feof(FileHandle) && (!found)) {
            if(fgets(line, linebuffer_size, FileHandle)) {
               if (strlen(line) > 32) {
                  line[32] = '\0'; // force it to 32 bytes
                  if (BN_hex2bn(&key_bn, line) != 32) {
                     (*rdprintf)("Invalid Processing Key / Device Key, ignoring\n");
                  } else {
                     BN_bn2bin(key_bn, key_bin);

                     memcpy(possible_keys + (keys_tried % 32) * 16, key_bin, 16); // store last 32 keys
                     if (keys_tried >= 32) keys_tried = 0;

                     keys_tried++;
                     //output_key(key_bin, 16, "Possible key", align, 0);

                     //number_of_uvs = 1;

                     // Try each c-value (and its corresponding uv)
                     for (int j = 0; (j < number_of_uvs) && (!found); j++) {
                        // get the c value
                        memcpy(encrypted_c_value, encrypted_c_values + 4 + j * 16, 16);
                        //(*rdprintf)("%d ", j);
                        //output_key(encrypted_c_value, 16, "Encrypted C-value", align, false);

                        // get the corresponding uv
                        memcpy(uv, uvs + 4 + 1 + j * 5, 4);

                        // get the u_mask (encoded as a number)
                        const unsigned char *u_mask_nr = uvs + 4 + j * 5;

                        // Check if any of the keys is the processing key we need
                        memcpy (processing_key, key_bin, 16);

                        if(check_processing_key()) {
                           found = 1;
                        }

                        if (!found) {
                           // Check if any of the keys is the device key we need
                           memcpy (device_key, key_bin, 16);

                           //(*rdprintf)("get_possible_processing_keys j = %d\n", j);
                           int nr_of_possible_processing_keys = get_possible_processing_keys(device_key, *u_mask_nr);

                           for (int k = 0; (k < nr_of_possible_processing_keys) && (!found); k++) {
                              memcpy(processing_key, possible_processing_keys + k * 16, 16);

                              //   output_key(processing_key, 16, "Possible proc key", align, 0);

                              if(check_processing_key()) {
                                 found = 1;
                                 if (verbose) {
                                    output_key(device_key, 16, "Device key", align, false);
                                 }
                              }
                           }
                        }

                     } // End For each c-value
                  }
               }
            }
         } // End while reading file
         fclose(FileHandle);
         if (!found) {
            (*rdprintf)("Could not find a Processing Key or Device Key resulting in the Media Key.\n");

            if (keys_tried >= 32) keys_tried = 32;
            for (int j = 0; j < keys_tried; j++) {
               output_key(possible_keys + j * 16, 16, "Possible key tried", align, false);
            }
            found = -3;
         }
      }
   }

   if (key_bn) {
      BN_free(key_bn);
   }

   return found;
}


int LoadMediaKeyFile(const char *dir) {
   // For building path to file
   char filename[_MAX_PATH];
   // Filepointer
   FILE *FileHandle = NULL;
   // For reading lines from the file
   const size_t linebuffer_size = 1024;
   char linebuffer[linebuffer_size];
   // Used for converting from hex string to binary
   BIGNUM *raw = NULL;
   // Result of conversion
   int rv = 0;
   // Return code
   int ret = 1;

   strcpy(filename, dir);
   strcat(filename, pathSeparator);
   strcat(filename, "MediaKey.txt");
   FileHandle = fopen(filename, "r");
   if (!FileHandle) {
      //(*rdprintf)("Could not open file: %s\n", filename);
      ret = -1;
   } else {
      // Read the Media Key
      if (fgets(linebuffer, linebuffer_size, FileHandle) != linebuffer) {
         (*rdprintf)("Could not read Media Key from file.\n");
         ret = -2;
      } else {
         // Limit the line length to 32 characters (== the 16 bytes of the Media Key)
         if (strlen(linebuffer) > 32) {
            linebuffer[32] = '\0';
         }
         rv = BN_hex2bn(&raw, linebuffer);
         if (rv != 32) {
            (*rdprintf)("Invalid Media Key in file MediaKey.txt\n");
            ret = -3;
         } else {
            BN_bn2bin(raw, media_key);
         }
      }
   }
   return ret;
}


int LoadHostKeyCertificate(const char *dir, /*out*/ unsigned char *key, /*out*/ unsigned char *cert, int level ) {
   // For building path to file
   char filename[_MAX_PATH];
   // Filepointer
   FILE *FileHandle = NULL;
   // For reading lines from the file
   const size_t linebuffer_size = 1024;
   char linebuffer[linebuffer_size];
   // Used for converting from hex string to binary
   BIGNUM *raw = NULL;
   // Result of conversion
   int rv = 0;
   // Return code
   int ret = 1;

   switch ( level ) {
   case 0:
      // normal mode
      strcpy(filename, dir);
      strcat(filename, pathSeparator);
      strcat(filename, "HostKeyCertificate.txt");
      FileHandle = fopen(filename, "r");
      if (!FileHandle) {
         (*rdprintf)("Could not open file: %s\n", filename);
         ret = -1;
      } else {
         // Read the Host Private Key
         if (fgets(linebuffer, linebuffer_size, FileHandle) != linebuffer) {
            (*rdprintf)("Could not read Host Private Key from file.\n");
            ret = -2;
         } else {
            // Limit the line length to 40 characters (== the 20 bytes of the Host Private Key)
            if (strlen(linebuffer) > 40) {
               linebuffer[40] = '\0';
            }
            rv = BN_hex2bn(&raw, linebuffer);
            if (rv != 40) {
               (*rdprintf)("Invalid Host Private Key\n");
               //memset(key, 0, 20);
               ret = -3;
            } else {
               BN_bn2bin(raw, key);

               // Read the Host Certificate
               if (fgets(linebuffer, linebuffer_size, FileHandle) != linebuffer) {
                  (*rdprintf)("Could not read Host Certificate from file.\n");
                  ret = -4;
               } else {
                  // Limit the line length to 184 characters (== the 92 bytes of the Host Certificate)
                  if (strlen(linebuffer) > 184) {
                     linebuffer[184] = '\0';
                  }
                  rv = BN_hex2bn(&raw, linebuffer);
                  if (rv != 184) {
                     (*rdprintf)("Invalid Host Certificate\n");
                     //memset(cert, 0, 92);
                     ret = -5;
                  } else {
                     BN_bn2bin(raw, cert);
                  }
               }
            }
         }

         fclose(FileHandle);
         if (raw) {
            BN_free(raw);
         }
      }

      break;

   case 1:
      // d option: investigate HRL of Drive
      memcpy ( key, PD7_priv_key, 20 );
      memcpy ( cert, PD7_host_cert, 92 );
      break;
   case 2:
      // d option: investigate HRL of Drive
      memcpy ( key, ORG_priv_key, 20 );
      memcpy ( cert, ORG_host_cert, 92 );
      break;
   case 3:
      // d option: investigate HRL of Drive
      memcpy ( key, V20_priv_key, 20 );
      memcpy ( cert, V20_host_cert, 92 );
      break;
   case 4:
      // d option: investigate HRL of Drive
      memcpy ( key, PS3_priv_key, 20 );
      memcpy ( cert, PS3_host_cert, 92 );
      break;
   default:
      break;
   }

   return ret;
}


void printUsage() {
   #if defined(WINDOWS) || defined(_WIN32)
   (*rdprintf)("Usage: aacskeys [options] <drive letter> [volume id / binding nonce / LBA]\n");
   #else
   (*rdprintf)("Usage: aacskeys [options] <mountpath> [volume id / binding nonce]\n");
   #endif
   (*rdprintf)("\n");
   (*rdprintf)("Authentication options for retrieving the VID / BN (default: automatic):\n");
   (*rdprintf)("If a Volume ID / Binding Nonce is given these options are ignored\n");
   (*rdprintf)(" -a : Force AACS authentication\n");
   (*rdprintf)(" -b : Force AACS bypass\n");
   (*rdprintf)(" -x : Force XBox hack\n");
   (*rdprintf)(" --no-preinval  : Do not invalidate all AGIDs before requesting one\n");
   (*rdprintf)(" --pa-lba=<LBA> : Use <LBA> as Starting LBA of the Protected Area\n");
   (*rdprintf)("\n");
   (*rdprintf)("Decrypting options:\n");
   (*rdprintf)(" --prefer-sca : Decrypt SCA content of HD-DVD's with both\n");
   (*rdprintf)("                content types (default: ACA)\n");
   (*rdprintf)(" --dump-vid   : Dump only the Volume ID\n");
   (*rdprintf)("\n");
   (*rdprintf)("Output options (default: only DiscID and VUK):\n");
   (*rdprintf)(" -s : Sensitive, same as verbose but doesn't hide sensitive data\n");
   (*rdprintf)(" -v : Verbose, gives lot of keys and info, but hides sensitive data\n");
   (*rdprintf)("\n");
   (*rdprintf)("Other options:\n");
   (*rdprintf)(" --exclusive-io : Use exclusive access for low level I/O\n");
   (*rdprintf)("                  This is only used under Mac OS X and may require root rights\n");
   (*rdprintf)("\n");
   (*rdprintf)("\n");
   (*rdprintf)("Examples:\n");
   #if defined(WINDOWS) || defined(_WIN32)
   (*rdprintf)("aacskeys i\n");
   (*rdprintf)("aacskeys -va i\n");
   (*rdprintf)("aacskeys i 16800\n");
   (*rdprintf)("aacskeys i 40000918200608410020202020200000\n");
   #else
   (*rdprintf)("aacskeys /media/cdrom\n");
   (*rdprintf)("aacskeys -va /media/cdrom\n");
   (*rdprintf)("aacskeys /media/cdrom 40000918200608410020202020200000\n");
   #endif
   #if defined(WINDOWS) || defined(_WIN32)
   (*rdprintf)("\n  or\n\n");
   (*rdprintf)("Usage: aacskeys <drive letter> [view types] <volume id / binding nonce / LBA>\n");
   (*rdprintf)("\n");
   (*rdprintf)("Possible view types (optional):\n");
   (*rdprintf)(" v = verbose (gives a lot of keys and info)\n");
   (*rdprintf)(" s = sensitive (gives a lot of keys and info AND gives potential sensitive drive\n");
   (*rdprintf)(" d = drive check (gives partial MKB version info)\n");
   (*rdprintf)(" w = write (CPSKey.cfg for BDAV)\n");
   (*rdprintf)("\n");
   (*rdprintf)("Examples:\n");
   (*rdprintf)("\n");
   (*rdprintf)("aacskeys i s 16800\n");
   (*rdprintf)("aacskeys i w 16800\n");
   (*rdprintf)("aacskeys i v 40000918200608410020202020200000\n");
   (*rdprintf)("aacskeys i w 40000918200608410020202020200000\n");
   (*rdprintf)("aacskeys i d\n");
   #endif
}


int main(int argc, char* argv[]) {
   int errnr = 0;
   char errtype[128];
   int returnvalue = EXIT_SUCCESS;

   enum authTypes { AUTOMATIC = 0, EXTERNAL, AACSAUTH, AACSBYPASS, XBOX };
   authTypes authType = AUTOMATIC;
   Drive drive;
   MMC mmc(&drive);
   unsigned char agid = 0;
   unsigned char bn_block_count = 0;
   char drive_path[_MAX_PATH];
   char drive_info[128];
   char aacsdir[_MAX_PATH];
   char tkfpath[_MAX_PATH];

   bool bluray = false;
   bool recordable = false;
   
   bool external_pa_lba = false;
   bool preinval = true;

   bool preferSca = false;
   bool dumpvid = false;

   bool verbose = false;
   bool hide_private_info = true;

   bool exclusiveIO = false;

   bool HRL_check = false;

   char textbuffer[256];
   int align = 32;
   int err_stat = 0;

   ERR_load_crypto_strings();

   // TODO: Is this necessary for all of these arrays?
   memset(errtype, 0, 128);
   memset(drive_path, 0, _MAX_PATH);
   memset(drive_info, 0, 128);
   memset(aacsdir, 0, _MAX_PATH);
   memset(tkfpath, 0, _MAX_PATH);
   memset(textbuffer, 0, 256);

   (*rdprintf)("aacskeys 0.4.0e by arnezami, KenD00, Key, Nobu1789\n\n");

   // Very simple command line parser
   if (argc < 2) {
      printUsage();
      return 1;
   }
   int currentArg = 1;
   // Parse options
   for (; currentArg < argc; currentArg++) {
      if (argv[currentArg][0] == '-') {
         if (strlen(argv[currentArg]) < 2) {
            (*rdprintf)("Unknown option: \"\"\n");
            return 1;
         }
         if (argv[currentArg][1] == '-') {
            // Long option
            if (strlen(argv[currentArg]) < 3) {
               (*rdprintf)("Unknown long option: \"\"\n");
               return 1;
            }
            char *lopt = argv[currentArg] + 2;
            if (strcmp(lopt, "no-preinval") == 0) {
               preinval = false;
            } else if (strncmp(lopt, "pa-lba=", 7) == 0) {
               char *endptr = NULL;
               unsigned long tmp = strtoul(lopt + 7, &endptr, 10);
               // strtoul accepts a leading minus
               if (*endptr == '\0' && lopt[7] != '-') {
                  pa_lba = tmp;
                  external_pa_lba = true;
               } else {
                  (*rdprintf)("Invalid Protected Area LBA, it must be an unsinged integer value\n");
                  return 1;
               }
            } else if (strcmp(lopt, "prefer-sca") == 0) {
               preferSca = true;
            } else if (strcmp(lopt, "dump-vid") == 0) {
               dumpvid = true;
            } else if (strcmp(lopt, "exclusive-io") == 0) {
               exclusiveIO = true;
            } else {
               (*rdprintf)("Unknown long option: \"%s\"\n", lopt);
               return 1;
            }
         } else {
            // Short option
            for (unsigned int currentOpt = 1; currentOpt < strlen(argv[currentArg]); currentOpt++) {
               switch (argv[currentArg][currentOpt]) {
               case 'a':
                  authType = AACSAUTH;
                  break;
               case 'b':
                  authType = AACSBYPASS;
                  break;
               case 'x':
                  authType = XBOX;
                  break;
               case 'v':
                  verbose = true;
                  hide_private_info = true;
                  break;
               case 's':
                  verbose = true;
                  hide_private_info = false;
                  break;
               default:
                  (*rdprintf)("Unknown option: \"%c\"\n", argv[currentArg][currentOpt]);
                  return 1;
               }
            }
         }
      } else {
         // Not an option
         break;
      }
   }
   // Parse drive
   if (currentArg == argc) {
      #if defined(WINDOWS) || defined(_WIN32)
      (*rdprintf)("No drive letter specified\n");
      #else
      (*rdprintf)("No mountpath specified\n");
      #endif
      return 1;
   }
   strncpy(drive_path, argv[currentArg++], _MAX_PATH - 1);
   #if defined(WINDOWS) || defined(_WIN32)
   // For windows, make drive of the form <driveletter>:
   drive_path[1] = ':';
   drive_path[2] = '\0';
   #else
   // Under *nix, cut off trailing "/"
   // Set the runner at the end of the string
   char *ptr = strrchr(drive_path, '\0');
   // Go one step before the terminating 0
   ptr--;
   // Decrement the runner until we find the first non / or hit the beginning of the string
   while (ptr >= drive_path && *ptr == '/') {
      // Cut off the path, if the path is root, the path will be empty
      *ptr = '\0';
      ptr--;
   }
   #endif
   if (currentArg < argc) {
       if (strlen(argv[currentArg]) < 2) {
         //Maybe View Type
         switch (argv[currentArg][0]) {
         case 'w':
            break;
         case 'd':
            HRL_check = true;
            break;
         case 'v':
            verbose = true;
            hide_private_info = true;
            break;
         case 's':
            verbose = true;
            hide_private_info = false;
            break;
         default:
            (*rdprintf)("Unknown option: \"%c\"\n", argv[currentArg][0]);
            return 1;
         }
         currentArg++;
      }
      if (currentArg < argc && strlen(argv[currentArg]) > 31) {
         // Parse VolumeID / Binding Nonce
         BIGNUM *volume_id_bn = NULL;
         int rv = BN_hex2bn(&volume_id_bn, argv[currentArg]);
         if (rv == 32) {
            BN_bn2bin(volume_id_bn, volume_id);
            BN_free(volume_id_bn);
         } else {
            BN_free(volume_id_bn);
            (*rdprintf)("Invalid Volume ID / Binding Nonce, it may contain only hexadecimal\ndigits (0-9, A-F) and must be 32 characters long\n");
            return 1;
         }
         authType = EXTERNAL;
         currentArg++;
      }
      if (currentArg < argc) {
         // Parse LBA
         char *endptr = NULL;
         unsigned long tmp = strtoul(argv[currentArg], &endptr, 10);
         // strtoul accepts a leading minus
         if (*endptr == '\0' && argv[currentArg][0] != '-') {
            pa_lba = tmp;
            external_pa_lba = true;
         } else {
            (*rdprintf)("Invalid Protected Area LBA, it must be an unsinged integer value\n");
            return 1;
         }
         currentArg++;
      }
   }
   // Check for remaining, invalid stuff
   if (currentArg < argc) {
      (*rdprintf)("Invalid parameter: \"%s\"\n", argv[currentArg]);
      return 1;
   }
   // End command line parser

   GetCurrentPath(currentpath);
   if (verbose) {
      output_text(currentpath, "Current path", align);
      (*rdprintf)("\n");
   }

   if (!hide_private_info) {
      (*rdprintf)("WARNING: the following contains sensitive information about your drive.\nDO NOT publish this on a forum or anywhere else on the internet.\nOtherwise your drive may be revoked!\n\n");
   }

   if ( !HRL_check ) {
      errnr = LoadMediaKeyBlockFile(drive_path, &bluray, &recordable, aacsdir);
      if (errnr < 0) {
         returnvalue = EXIT_FAILURE;
         strcpy(errtype, "LOADMKB");
         goto exit;
      }
   }

   if (bluray) {
      mmc.setMediaType(MMC::BLURAY);
   }

   if (verbose) {
      const unsigned char *tvRecord = GetRecordFromFile(MediaKeyFileBuffer, FileSizeMKF, 0x10);
      if (tvRecord != NULL) {
         unsigned int mkbv = tvRecord[8];
         mkbv <<= 8;
         mkbv |= tvRecord[9];
         mkbv <<= 8;
         mkbv |= tvRecord[10];
         mkbv <<= 8;
         mkbv |= tvRecord[11];
         sprintf(textbuffer, "%d", mkbv);
      } else {
         strcpy(textbuffer, "Type And Version record not found!");
      }
      output_text(textbuffer, "MKBv", align);
   }

   if ( !HRL_check ) {
      errnr = LoadTitleKeyFile(drive_path, bluray, recordable, preferSca, tkfpath);
      if (errnr < 0) {
         returnvalue = EXIT_FAILURE;
         strcpy(errtype, "LOADTKF");
         goto exit;
      }
   }


   if (!dumpvid && !HRL_check) {
      // This function processes the Media Key Block (trying all possible Processing/Device Keys)
      errnr = LoadMediaKeyFile(currentpath);
      if ( errnr < 0 ) {
         errnr = process_media_key_block(align, verbose);
         if (errnr < 0) {
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "PROCESSMKB");
            goto exit;
         }
         if (verbose) {
            output_key(processing_key, 16, "Processing key", align, false);
         }

         // Decrypt the c value into the media key (using the processing key and the corresponding uv)
         decrypt_c_value(encrypted_c_value, processing_key, decrypted_c_value, uv, media_key);
         if (verbose) {
            output_key(encrypted_c_value, 16, "Encrypted C-value", align, false);
             output_key(uv, 4, "Corresponding uv", align, false);
             (*rdprintf)("\n");
             output_key(decrypted_c_value, 16, "Decrypted C-value", align, false);
         }
      } else {
         (*rdprintf)("Got Media Key from file. No process MKB.\n");
      }

      if (verbose) {
         output_key(media_key, 16, "Media key", align, false);
      }

      // Decrypt the verification data (using the media key)
      decrypt_verification_data(encrypted_verification_data, media_key, decrypted_verification_data);
      if (verbose) {
         (*rdprintf)("\n");
         output_key(encrypted_verification_data, 16, "Encrypted verification data", align, false);
         output_key(decrypted_verification_data_should_be, 8, "Decr verif data should be", align, false);
         output_key(decrypted_verification_data, 16, "Decrypted verification data", align, false);
         (*rdprintf)("\n");
      }
      if (memcmp(decrypted_verification_data_should_be, decrypted_verification_data, 8)) {
         output_key(media_key, 16, "Invalid Media key", align, false);
         returnvalue = EXIT_FAILURE;
         strcpy(errtype, "VERIFYDATA");
         errnr = -1;
         goto exit;
      }
   }

   // ###DEBUG###
   //recordable = true;

   if (authType != EXTERNAL) {
      bool readVid = false;
      errnr = drive.open(drive_path, exclusiveIO);
      if (errnr < 0) {
         if (errnr == -1) {
            (*rdprintf)("Could not open drive, maybe no media is inserted, aborting.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "DRIVEOPEN");
            goto exit;
         } else if (errnr == -2) {
            (*rdprintf)("Could not open drive, invalid drive type or mount device could not be found, aborting.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "DRIVETYPEMOUNT");
            goto exit;
         } else if (errnr == -3) {
            (*rdprintf)("Could not open /proc/mounts for reading, aborting.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "DRIVEMOUNTS");
            goto exit;
         } else if (errnr == -4) {
            (*rdprintf)("Could not follow symlink, aborting.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "DRIVESYMLINK");
            goto exit;
         } else if (errnr == -5) {
            (*rdprintf)("Could not get IOService, aborting.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "DRIVEIOSERVICE");
            goto exit;
         } else if (errnr == -6) {
            (*rdprintf)("Could not get MMCPlugIn, aborting.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "DRIVEMMCPLUGIN");
            goto exit;
         } else if (errnr == -7) {
            (*rdprintf)("Could not get MMCInterface, aborting.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "DRIVEMMCINTERFACE");
            goto exit;
         } else if (errnr == -8) {
            (*rdprintf)("Could not get SCSITaskDeviceInterface, aborting.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "DRIVESCSITASKDEVICEINTERFACE");
            goto exit;
         } else if (errnr == -9) {
            (*rdprintf)("Could not get exclusive access, aborting.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "DRIVEEXCLUSIVE");
            goto exit;
         } else {
            (*rdprintf)("Unknown error, aborting.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "DRIVEUNKNOWN");
            goto exit;
         }
      }

      errnr = mmc.read_drive_info(drive_info);
      if (errnr < 0) {
         (*rdprintf)("Error reading drive information.\n");
         returnvalue = EXIT_FAILURE;
         strcpy(errtype, "READDRIVEINFO");
         if (errnr == -2) {
            strcat(errtype, ": ");
            strcat(errtype, drive.sense_cmd().c_str());
         }
         goto err_close;
      }
      if (verbose) {
         output_text(drive_info, "Drive FW info", align);
         #ifdef DEBUG
         output_key(reinterpret_cast<unsigned char*>(drive_info), strlen(drive_info), "Drive FW info raw", align, false);
         #endif
      }

      unsigned char aacs_version = 0;
      unsigned char concurrent_agids = 0;
      bool supports_bng = false;
      bool feature_enabled = false;
      errnr = mmc.read_aacs_feature(&aacs_version, &concurrent_agids, &supports_bng, &bn_block_count, &feature_enabled);
      if (errnr < 0) {
         (*rdprintf)("Error reading AACS Feature Descriptor.\n");
         returnvalue = EXIT_FAILURE;
         strcpy(errtype, "GETAACSFEATURE");
         if (errnr == -2) {
            strcat(errtype, ": ");
            strcat(errtype, drive.sense_cmd().c_str());
         }
         goto err_close;
      }
      if (verbose) {
         output_key(&aacs_version, sizeof(aacs_version), "AACS Version", align, false);
         sprintf(textbuffer, "%d", concurrent_agids);
         output_text(textbuffer, "Number of concurrent AGIDs", align);
         if (supports_bng) {
            sprintf(textbuffer, "YES");
         } else {
            sprintf(textbuffer, "NO");
         }
         output_text(textbuffer, "Supports BN generation", align);
         sprintf(textbuffer, "%d", bn_block_count);
         output_text(textbuffer, "BN Block Count", align);
         if (feature_enabled) {
            sprintf(textbuffer, "YES");
         } else {
            sprintf(textbuffer, "NO");
         }
         output_text(textbuffer, "Inserted medium AACS protected", align);
         (*rdprintf)("\n");
      }

      if (recordable) {
         // TODO: Does the bn_block_count needs to be adjusted?
         // The XBox-Drive reports 0, but the HD spec says the BN spans 4 sectors
         // The GGW-H20L reports 4 for HD and 1 for BD, the BD spec doesn't say how many sectors the BN spans
         // The AACS spec says only that this number is the minimum!
         if (bluray) {
            if (bn_block_count < 1) {
                bn_block_count = 1;
            }
         } else {
            if (bn_block_count < 4) {
               bn_block_count = 4;
            }
         }

         // FIXME: Implement UDF parsing to get the LBA if not specified, maybe this should be done after the XBox hack?
         if (!external_pa_lba && !HRL_check) {
            errnr = -1;
            (*rdprintf)("Error, UDF parsing not implemented and no Starting LBA specified.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "NOPALBA");
            goto err_close;
         }
      }

      if (((authType == AUTOMATIC) && (!strcmp(drive_info, "MC0810/03/06") || !strcmp(drive_info, "TS0602/26/07"))) || (authType == XBOX)) {
      // ###DEBUG###
      /* if (((authType == AUTOMATIC) && (!strcmp(drive_info, "MC0810/03/06"))) || (authType == XBOX)) */
         if (!recordable) {
            errnr = mmc.read_vid_mc08(0, volume_id, mac);
            if (errnr < 0) {
               (*rdprintf)("Error reading XBOX VID.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "READVIDXBOX");
               if (errnr == -2) {
                  strcat(errtype, ": ");
                  strcat(errtype, drive.sense_cmd().c_str());
               }
               goto err_close;
            }
            readVid = true;
            if (verbose || dumpvid) {
               output_key(volume_id, 16, "Volume ID (XBOX)", align, false);
               (*rdprintf)("\n");
            }
         } else {
            errnr = -1;
            (*rdprintf)("Error, XBox hack not implemented for reading Binding Nonces.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "READBNXBOX");
            goto err_close;
         }
      }

      if (((authType == AUTOMATIC) && !readVid) || (authType == AACSBYPASS)) {
         // Invalidate all AGIDs, just in case someone left some used
         if (preinval) {
            mmc.invalidate_agid(0);
            mmc.invalidate_agid(1);
            mmc.invalidate_agid(2);
            mmc.invalidate_agid(3);
         }
         errnr = mmc.report_agid(&agid);
         if (errnr < 0) {
            (*rdprintf)("All AGIDs in use.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "AGID");
            if (errnr == -2) {
               strcat(errtype, ": ");
               strcat(errtype, drive.sense_cmd().c_str());
            }
            goto err_close;
         }
         if (verbose) {
            output_key(&agid, 1, "AGID", align, false);
            (*rdprintf)("\n");
         }

         if (!recordable) {
            errnr = mmc.read_vid(agid, volume_id, mac);
            if (errnr < 0) {
               // In case of forced AACSBYPASS produce an error, in automatic mode ignore it and advance
               if (authType == AACSBYPASS) {
                  (*rdprintf)("Error reading AACSBYPASS VID.\n");
                  returnvalue = EXIT_FAILURE;
                  strcpy(errtype, "READVIDAACSBYPASS");
                  if (errnr == -2) {
                     strcat(errtype, ": ");
                     strcat(errtype, drive.sense_cmd().c_str());
                  }
                  goto err_invalidate;
               }
            } else {
               readVid = true;
               if (verbose || dumpvid) {
                  output_key(volume_id, 16, "Volume ID (AACSBYPASS)", align, false);
                  (*rdprintf)("\n");
               }
            }
         } else {
            errnr = mmc.read_bn(agid, pa_lba, bn_block_count, volume_id, mac);
            if (errnr < 0) {
               // In case of forced AACSBYPASS produce an error, in automatic mode ignore it and advance
               if (authType == AACSBYPASS) {
                  (*rdprintf)("Error reading AACSBYPASS BN.\n");
                  returnvalue = EXIT_FAILURE;
                  strcpy(errtype, "READBNAACSBYPASS");
                  if (errnr == -2) {
                     strcat(errtype, ": ");
                     strcat(errtype, drive.sense_cmd().c_str());
                  }
                  goto err_invalidate;
               }
            } else {
               readVid = true;
               if (verbose) {
                  output_key(volume_id, 16, "Binding Nonce (AACSBYPASS)", align, false);
                  (*rdprintf)("\n");
               }
            }
         }
         // This is reached if an error in automatic mode happened or the VID was read, need to invalidate the AGID now
         // because a new one gets requested in the next automatic step or it won't get invalidated later.
         mmc.invalidate_agid(agid);
      }

      if (((authType == AUTOMATIC) && !readVid) || (authType == AACSAUTH)) {
         #ifndef AACS_NO_ECDSA
         // Read Host Private Key and Host Certificate from file

         int level;
         int start;
         int max;
         if ( !HRL_check ) {
             // no check HRL,  Try 1 Host Private Key, default or read from file
             max = 1;
             start = 0;
         } else {
             // check HRL,  Try 4 Host Private Keys
             max = 5;
             start = 1;
         }

         for ( level = start; level < max; level++ )  {
            err_stat = 0;
            errnr = LoadHostKeyCertificate(currentpath, priv_key, host_cert, level);
            if (errnr < 0 && !HRL_check ) {
               (*rdprintf)("Error reading Host Private Key and Host Certificate from file.\n");
               //returnvalue = EXIT_FAILURE;
               //strcpy(errtype, "READHOSTKEYCERT");
               //goto err_close;
               (*rdprintf)("Use Default Host Private Key and Host Certificate.\n");
            }
            if (verbose) {
               output_key(priv_key, 20, "Host Private Key (Hpriv)", align, false);
               output_key(host_cert, 92, "Host certificate (Hcert)", align, false);
               output_key(host_nonce, 20, "Host Nonce (Hn)", align, false);
               (*rdprintf)("\n");
            }
   
            // Invalidate all AGIDs, just in case someone left some used
            if (preinval) {
               mmc.invalidate_agid(0);
               mmc.invalidate_agid(1);
               mmc.invalidate_agid(2);
               mmc.invalidate_agid(3);
            }
            errnr = mmc.report_agid(&agid);
            if (errnr < 0) {
               if ( !HRL_check ) (*rdprintf)("All AGIDs in use.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "AGID");
               if (errnr == -2) {
                  strcat(errtype, ": ");
                  strcat(errtype, drive.sense_cmd().c_str());
               }
               err_stat = 1;
               continue;
               //goto err_close;
            }
            if (verbose) {
               output_key(&agid, 1, "AGID", align, false);
               (*rdprintf)("\n");
            }
   
            errnr = mmc.send_host_cert_chal(agid, host_nonce, host_cert);
            if (errnr < 0) {
               if ( !HRL_check ) (*rdprintf)("The given Host Certficate / Private Key has been revoked by your drive.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "SENDHOSTCHAL");
               if (errnr == -2) {
                  strcat(errtype, ": ");
                  strcat(errtype, drive.sense_cmd().c_str());
               }
               err_stat = 2;
               continue;
               //goto err_invalidate;
            }
   
            errnr = mmc.report_drive_cert_chal(agid, drive_nonce, drive_cert);
            if (errnr < 0) {
               if ( !HRL_check ) (*rdprintf)("The drive isn't giving back its Drive Certificate.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "REPORTDRIVECHAL");
               if (errnr == -2) {
                  strcat(errtype, ": ");
                  strcat(errtype, drive.sense_cmd().c_str());
               }
               err_stat = 2;
               continue;
               //goto err_invalidate;
            }
            if (verbose) {
               output_key(drive_cert, 92, "Drive certificate (Dcert)", align, hide_private_info);
               output_key(drive_nonce, 20, "Drive Nonce (Dn)", align, hide_private_info);
               (*rdprintf)("\n");
            }
   
            errnr = mmc.report_drive_key(agid, drive_key_point, drive_key_sig);
            if (errnr < 0) {
               if ( !HRL_check ) (*rdprintf)("The drive isn't giving back its Drive Key.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "REPORTDRIVEKEY");
               if (errnr == -2) {
                  strcat(errtype, ": ");
                  strcat(errtype, drive.sense_cmd().c_str());
               }
               err_stat = 2;
               continue;
               //goto err_invalidate;
            }
            if (verbose) {
               output_key(drive_key_point, 40, "Drive key point (Dv)", align, hide_private_info);
               output_key(drive_key_sig, 40, "Drive key signature (Dsig)", align, hide_private_info);
               (*rdprintf)("\n");
            }
   
            errnr = aacs_sign(host_cert, priv_key, host_key_sig, drive_nonce, host_key_point);
            if (errnr < 0) {
               if ( !HRL_check ) (*rdprintf)("Problem with creating a Host key signature.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "AACSSIGN");
               err_stat = 2;
               continue;
               //goto err_invalidate;
            }
            if (verbose) {
               output_key(host_key, 20, "Host key (Hk)", align, false);
               output_key(host_key_point, 40, "Host key point (Hv)", align, false);
               output_key(host_key_sig, 40, "Host key signature (Hsig)", align, hide_private_info);
               (*rdprintf)("\n");
            }
   
            errnr = aacs_verify(drive_cert, drive_key_sig, host_nonce, drive_key_point);
            if (errnr < 0) {
               if ( !HRL_check ) (*rdprintf)("Problem with verifying the drive signature.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "AACSVERYIFDRIVE");
               err_stat = 2;
               continue;
               //goto err_invalidate;
            }
   
            errnr = aacs_verify(host_cert, host_key_sig, drive_nonce, host_key_point);
            if (errnr < 0) {
               if ( !HRL_check ) (*rdprintf)("Problem with verifying the host signature.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "AACSVERYIFHOST");
               err_stat = 2;
               continue;
               //goto err_invalidate;
            }
   
            errnr = mmc.send_host_key(agid, host_key_point, host_key_sig);
            if (errnr < 0) {
               if ( !HRL_check ) (*rdprintf)("Error sending the Host signature and Key point.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "SENDHOSTKEY");
               if (errnr == -2) {
                  strcat(errtype, ": ");
                  strcat(errtype, drive.sense_cmd().c_str());
               }
               err_stat = 2;
               continue;
               //goto err_invalidate;
            }
            break;

         }

         if ( HRL_check ) {
             if ( err_stat == 0 ) {
               // found some valid Host Private Key
               switch ( level ) {
               case 1:
                  (*rdprintf)("Drive Partial MKB = V1\n");
                  break;
               case 2:
                  (*rdprintf)("V1 < Drive Partial MKB < V17\n");
                  break;
               case 3:
                  (*rdprintf)("V16 < Drive Partial MKB < V21\n");
                  break;
               case 4:
                  (*rdprintf)("V20 < Drive Partial MKB < V29\n");
                  break;
               default:
                  break;
               }
             } else {
               // all Host Private Keys revoked
               (*rdprintf)("Drive Partial MKB > V28\n");
             }
             errnr = 0;
             goto exit;
         } else {
            if ( err_stat == 1 ) {
               goto err_close;
            } else if ( err_stat == 2 ) {
               goto err_invalidate;
            }
         }
   
         errnr = aacs_calculate_bus_key(host_cert, host_key, drive_key_point, bus_key);
         if (errnr < 0) {
            (*rdprintf)("Problem with calculating Bus Key.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "CALCBUSKEY");
            goto err_invalidate;
         }
         if (verbose) {
            output_key(bus_key, 16, "Bus key (BK)", align, hide_private_info);
            (*rdprintf)("\n");
         }

         if (!recordable) {
            errnr = mmc.read_vid(agid, volume_id, mac);
            if (errnr < 0) {
               (*rdprintf)("Error reading AACSAUTH VID.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "READVIDAACSAUTH");
               if (errnr == -2) {
                  strcat(errtype, ": ");
                  strcat(errtype, drive.sense_cmd().c_str());
               }
               goto err_invalidate;
            }
            readVid = true;
            if (verbose || dumpvid) {
               output_key(volume_id, 16, "Volume ID (AACSAUTH)", align, false);
            }
         } else {
            errnr = mmc.read_bn(agid, pa_lba, bn_block_count, volume_id, mac);
            if (errnr < 0) {
               (*rdprintf)("Error reading AACSAUTH BN.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "READBNAACSAUTH");
               if (errnr == -2) {
                  strcat(errtype, ": ");
                  strcat(errtype, drive.sense_cmd().c_str());
               }
               goto err_invalidate;
            }
            readVid = true;
            if (verbose || dumpvid) {
               output_key(volume_id, 16, "Binding Nonce (AACSAUTH)", align, false);
            }
         }

         // The AGID isn't used later any more, invalidate it right here.
         mmc.invalidate_agid(agid);
         // The drive isn't used any more, close it here
         drive.close();

         if (!recordable) {
            errnr = calculate_volume_id_mac(volume_id, 16, bus_key, volume_id_mac);
            if (errnr < 0) {
               (*rdprintf)("Problem with calculating Volume ID MAC.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "CALCVIDMAC");
               goto exit;
            }
            if (verbose) {
               output_key(volume_id_mac, 16, "Volume ID MAC", align, hide_private_info);
               output_key(mac, 16, "Volume ID MAC should be", align, hide_private_info);
               (*rdprintf)("\n");
            }
            if (memcmp(volume_id_mac, mac, 16)) {
               (*rdprintf)("The VID MAC is incorrect. This means this Volume ID is not correct.\n");
               returnvalue = EXIT_FAILURE;
               errnr = -1;
               strcpy(errtype, "VIDMAC");
               goto exit;
            }
         } else {
            // Append the Protected Area LBA to the Binding Nonce to verify the MAC
            unsigned char bn_lba[20];
            memcpy(bn_lba, volume_id, 16);
            bn_lba[16] = (pa_lba >> 24) & 0xFF;
            bn_lba[17] = (pa_lba >> 16) & 0xFF;
            bn_lba[18] = (pa_lba >> 8) & 0xFF;
            bn_lba[19] = pa_lba & 0xFF;
            errnr = calculate_volume_id_mac(bn_lba, 20, bus_key, volume_id_mac);
            if (errnr < 0) {
               (*rdprintf)("Problem with calculating Binding Nonce MAC.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "CALCBNMAC");
               goto exit;
            }
            if (verbose) {
               output_key(volume_id_mac, 16, "Binding Nonce MAC", align, hide_private_info);
               output_key(mac, 16, "Binding Nonce MAC should be", align, hide_private_info);
               (*rdprintf)("\n");
            }
            if (memcmp(volume_id_mac, mac, 16)) {
               errnr = -1;
               (*rdprintf)("The BN MAC is incorrect. This means this Binding Nonce is not correct.\n");
               returnvalue = EXIT_FAILURE;
               strcpy(errtype, "BNMAC");
               goto exit;
            }
         }
         #else
         errnr = -1;
         (*rdprintf)("AACSAUTH support not compiled.\n");
         returnvalue = EXIT_FAILURE;
         strcpy(errtype, "NOAACSAUTH");
         goto err_close;
         #endif
      }
   } else {
      if (verbose || dumpvid) {
         if (!recordable) {
            output_key(volume_id, 16, "Volume ID (EXTERNAL)", align, false);
         } else {
            output_key(volume_id, 16, "Binding Nonce (EXTERNAL)", align, false);
         }
         (*rdprintf)("\n");
      }
   }

   // ###DEBUG###
   //recordable = true;

   if (!dumpvid) {
      calculate_volume_unique_key(volume_id, media_key, decrypted_volumeid, volume_unique_key);
      if (!recordable) {
         output_key(volume_unique_key, 16, "Volume Unique Key", align, false);
      } else {
         output_key(volume_unique_key, 16, "Protected Area Key", align, false);
      }
   }

   calculate_title_key_file_hash(TitleKeyFileBuffer, FileSizeTKF, title_key_file_hash);

   if (!dumpvid) {
      if (!bluray) {
         output_key(title_key_file_hash, 20, "TKF Hash (Disc ID)", align, false);

         // TODO: This part needs to be fixed for recordables
         // global FileSize...
         calculate_title_key_file_mac(TitleKeyFileBuffer, FileSizeTKF - 16, volume_unique_key, title_key_file_mac_calc);
         if (verbose) {
            output_key(title_key_file_mac_calc, 16, "Title Key File MAC", align, false);
            output_key(title_key_file_mac, 16, "TKF MAC should be", align, false);
            (*rdprintf)("\n");
         }
         if (memcmp(title_key_file_mac_calc, title_key_file_mac, 16)) {
            errnr = -1;
            (*rdprintf)("The VUK MAC is incorrect. This means this Volume Unique Key is not correct.\n");
            returnvalue = EXIT_FAILURE;
            strcpy(errtype, "VUKMAC");
            goto exit;
         }

         for (int i = 0; i < nr_of_title_keys; i++) {
            if (verbose) {
               sprintf(textbuffer, "Encrypted Title Key %d", i + 1);
               output_key(title_keys[i].encrypted_title_key, 16, textbuffer, align, false);
            }

            // Decrypt each title key/cps unit key (using the volume unqiue key)
            decrypt_title_key(title_keys[i].encrypted_title_key, volume_unique_key, NULL, 0, title_keys[i].decrypted_title_key);
         }
         if (verbose) {
            (*rdprintf)("\n");
            for (int i = 0; i < nr_of_title_keys; i++) {
               sprintf(textbuffer, "Decrypted Title Key %d", i+1);
               output_key(title_keys[i].decrypted_title_key, 16, textbuffer, align, false);
            }
         }
      } else {
         output_key(title_key_file_hash, 20, "Unit Key File Hash (Disc ID)", align, false);
         if (verbose) {
            (*rdprintf)("\n");
         }

         // Buffer for constructing the path to the Usage Rules file
         char ukfpath[_MAX_PATH];
         // Buffer to hold the Usage Rules file, initially NULL, the first realloc will initialize it
         size_t ukfsize = 0;
         unsigned char *ukfbuffer = NULL;

         for (int i = 0; i < nr_of_title_keys; i++) {
            if (verbose) {
               sprintf(textbuffer, "Encrypted Unit Key %d", i+1);
               output_key(title_keys[i].encrypted_title_key, 16, textbuffer, align, false);
            }

            // Decrypt each title key/cps unit key (using the volume unqiue key)
            if (!recordable) {
               decrypt_title_key(title_keys[i].encrypted_title_key, volume_unique_key, NULL, 0, title_keys[i].decrypted_title_key);
            } else {
               // Build path to Usage Rules file
               strcpy(ukfpath, aacsdir);
               strcat(ukfpath, pathSeparator);
               sprintf(ukfpath + strlen(ukfpath), "CPSUnit%05d.cci", i + 1);
               FILE *ukffp = fopen(ukfpath, "rb");
               if (ukffp != NULL) {
                  // Determine size of Usage Rules file
                  fseek(ukffp, 0, SEEK_END);
                  size_t new_ukfsize = ftell(ukffp);
                  rewind(ukffp);
                  // Resize buffer if necessary
                  if (ukfsize < new_ukfsize) {
                     unsigned char *temp = (unsigned char *) realloc(ukfbuffer, new_ukfsize);
                     if (temp != NULL) {
                        ukfbuffer = temp;
                        ukfsize = new_ukfsize;
                     } else {
                        // Set decrypted Unit Key to 0 to indicate an error
                        memset(title_keys[i].decrypted_title_key, 0, 16);
                        (*rdprintf)("Error decrypting Unit Key %d, could not (re)allocate memory for Usage Rules file.\n", i + 1);
                        // Skip the rest and continue with next loop
                        fclose(ukffp);
                        continue;
                     }
                  }
                  // Read the Usage Rules file into memory
                  if (fread(ukfbuffer, sizeof(unsigned char), ukfsize, ukffp) == ukfsize) {
                     decrypt_title_key(title_keys[i].encrypted_title_key, volume_unique_key, ukfbuffer, ukfsize, title_keys[i].decrypted_title_key);
                  } else {
                     // Set decrypted Unit Key to 0 to indicate an error
                     memset(title_keys[i].decrypted_title_key, 0, 16);
                     (*rdprintf)("Error decrypting Unit Key %d, could not fully read Usage Rules file.\n", i + 1);
                  }
                  fclose(ukffp);
               } else {
                  // Set decrypted Unit Key to 0 to indicate an error
                  memset(title_keys[i].decrypted_title_key, 0, 16);
                  (*rdprintf)("Error decrypting Unit Key %d, could not open Usage Rules file %s!\n", i + 1, ukfpath);
               }
            }
         }

         if (ukfbuffer) {
            free(ukfbuffer);
         }

         if (verbose) {
            (*rdprintf)("\n");
            for (int i = 0; i < nr_of_title_keys; i++) {
               sprintf(textbuffer, "Decrypted Unit Key %d", i + 1);
               output_key(title_keys[i].decrypted_title_key, 16, textbuffer, align, false);
            }
         }
      }
   }

   if (!hide_private_info) {
      (*rdprintf)("\nWARNING: the above contains sensitive information about your drive.\nDO NOT publish this on a forum or anywhere else on the internet.\nOtherwise your drive may be revoked!\n\n");
   }

   if (!verbose) {
      char CPSKeypath[_MAX_PATH];
      strcpy(CPSKeypath, currentpath);
      strcat(CPSKeypath, pathSeparator);
      sprintf(CPSKeypath + strlen(CPSKeypath), "CPSKey.cfg");
      FILE *CPSKeyfp = fopen(CPSKeypath, "wb");
      if (CPSKeyfp != NULL) {
         fprintf(CPSKeyfp, "; Key database file for BluRay(BDAV) version.\n");
         fprintf(CPSKeyfp, "0102030405060708090A0B0C0D0E0F1011121314=BDAV(");
         for (int i = 0; i < 16; i++) {
            fprintf(CPSKeyfp, "%02X", volume_id[i]);
         }
         fprintf(CPSKeyfp, ")|T|Mm/Dd/Yy");
         for (int i = 0; i < nr_of_title_keys; i++) {
            fprintf(CPSKeyfp, "|%2d-", i + 1);
            for (int j = 0; j < 16; j++) {
               fprintf(CPSKeyfp, "%02X", title_keys[i].decrypted_title_key[j]);
            }
         }
         fprintf(CPSKeyfp, "\n");
         fclose(CPSKeyfp);
      }
   }

   // *** JNI MOD start: set the retrieved keys (except Tile Keys, TODO: Include them as well?) to the keyData object via the global globalJni ***
   #ifdef JNILIB
   // Check globalJni. This is not really necessary except the loaded library is used by other code than the JNI methods
   if (globalJni != NULL) {
      // Verify that the DiscID's match
      // Have to cast here because jobject gets returned. Does dynamic cast work to ensure type safety?
      jbyteArray jdiscIdArray = static_cast<jbyteArray>(globalJni->env->CallObjectMethod(globalJni->keyData, globalJni->getDiscId));
      if (jdiscIdArray == NULL) {
         returnvalue = EXIT_FAILURE;
         errnr = -1;
         strcpy(errtype, "JNIGETDISCID");
         goto exit;
      }
      if (globalJni->env->GetArrayLength(jdiscIdArray) != 20) {
         returnvalue = EXIT_FAILURE;
         errnr = -2;
         strcpy(errtype, "JNIGETDISCID");
         goto exit;
      }
      jbyte jdiscId[20];
      globalJni->env->GetByteArrayRegion(jdiscIdArray, 0, 20, jdiscId);
      if (memcmp(title_key_file_hash, jdiscId, 20) != 0) {
         returnvalue = EXIT_FAILURE;
         errnr = -1;
         strcpy(errtype, "JNIDISCIDMISMATCH");
         goto exit;
      }

      // Create Java byte array that is used for all following operations
      jbyteArray jkey = globalJni->env->NewByteArray(16);
      if (jkey == NULL) {
         returnvalue = EXIT_FAILURE;
         errnr = -1;
         strcpy(errtype, "JNISETKEYS");
         goto exit;
      }

      // The following methods may all throw an ArrayIndexOutOfBoundsException,
      // but we assume that the key lengths are all correct and therefor don't check for exceptions
      if (!dumpvid) {
         // Set the Media Key
         globalJni->env->SetByteArrayRegion(jkey, 0, 16, reinterpret_cast<jbyte*>(media_key));
         globalJni->env->CallVoidMethod(globalJni->keyData, globalJni->setMek, jkey, 0);
      }
      // Set the Volume ID
      globalJni->env->SetByteArrayRegion(jkey, 0, 16, reinterpret_cast<jbyte*>(volume_id));
      globalJni->env->CallVoidMethod(globalJni->keyData, globalJni->setVid, jkey, 0);
      if (!dumpvid) {
         // Set the Volume Unique Key
         globalJni->env->SetByteArrayRegion(jkey, 0, 16, reinterpret_cast<jbyte*>(volume_unique_key));
         globalJni->env->CallVoidMethod(globalJni->keyData, globalJni->setVuk, jkey, 0);
         // Set all Title / CPS Unit Keys
         // TODO: Filter all 0-keys (== nonpresent Title Key)
         for (int i = 0; i < nr_of_title_keys; i++) {
            globalJni->env->SetByteArrayRegion(jkey, 0, 16, reinterpret_cast<jbyte*>(title_keys[i].decrypted_title_key));
            globalJni->env->CallVoidMethod(globalJni->keyData, globalJni->setTuk, i + 1, jkey, 0);
         }
      }
   }
   #endif
   // *** JNI MOD end ***

   // Overjump the exit targets for the drive part, if we reach this line the drive is already closed!
   goto exit;

err_invalidate:
   mmc.invalidate_agid(agid);
err_close:
   drive.close();
exit:
   if (errnr < 0) {
      (*rdprintf)("\n\nERROR: %s, errnr: %d\n\n", errtype, errnr);
      // *** JNI MOD start: Throw java side exception in case of error ***
      #ifdef JNILIB
      // Throw a java side exception
      if (globalJni != NULL) {
         // Misuse the buffer used by printfj(const char*, ...)
         sprintf(sprintfBuffer, "aacskeys ERROR: %s, errnr: %d", errtype, errnr);
         globalJni->env->ThrowNew(globalJni->aacsException, sprintfBuffer);
      }
      #endif
      // *** JNI MOD end ***
   }
   if (MediaKeyFileBuffer) {
      free(MediaKeyFileBuffer);
   }
   if (TitleKeyFileBuffer) {
      free(TitleKeyFileBuffer);
   }

   return returnvalue;
}



// *** JNI MOD start ***
#ifdef JNILIB
/*
* This is the target function for java side output of the function pointer
*/
int printfj(const char *fmt, ...) {
   int returnVal = 0;
   va_list args;
   va_start(args, fmt);
   // This is the safer version of vsprintf, is it contained in the standard?
   //vsnprintf(sprintfBuffer, 4096, fmt, args);
   returnVal = vsprintf(sprintfBuffer, fmt, args);
   va_end(args);
   //printf(sprintfBuffer);
   //fflush(stdout);
   // This may cause an exception, but this is rather unlikely
   globalJni->env->CallVoidMethod(globalJni->messagePrinter, globalJni->print, globalJni->env->NewStringUTF(sprintfBuffer));
   return returnVal;
}

/*
* Class:     dumphd_aacs_AACSKeys
* Method:    getVersionString
* Signature: ()Ljava/lang/String;
*/
JNIEXPORT jstring JNICALL Java_dumphd_aacs_AACSKeys_getVersionString
   (JNIEnv *env, jobject jobj) {
   // Return a string describing the library version
   return env->NewStringUTF("aacskeys library 0.4.0e by arnezami, KenD00, Key, Nobu1789");
}

/*
* Class:     dumphd_aacs_AACSKeys
* Method:    getKeys
* Signature: (Ljava/lang/String;Ldumphd/core/KeyData;)V
*/
JNIEXPORT void JNICALL Java_dumphd_aacs_AACSKeys_getKeys
   (JNIEnv *env, jobject jthis, jstring jpath, jobject jkeyData) {
   // *****************************************************
   // *** Prepare access to necessary jave side objects ***
   // *****************************************************
   JNI_MEMBERS jni_members;
   jni_members.env = env;
   jni_members.keyData = jkeyData;

   // Retrieve the AACSException class
   jni_members.aacsException = env->FindClass("dumphd/aacs/AACSException");
   if (jni_members.aacsException == NULL) {
      env->FatalError("Class dumphd.aacs.AACSException not found");
   }

   // Retrieve the messagePrinter object and its class from the java AACSKeys object
   jclass jthisClass = env->GetObjectClass(jthis);
   jfieldID mpFieldID = env->GetFieldID(jthisClass, "mp", "Ldumphd/util/MessagePrinter;");
   if (mpFieldID == NULL) {
      env->FatalError("Field AACSKeys->mp not found");
   }
   jni_members.messagePrinter = env->GetObjectField(jthis, mpFieldID);
   jclass jmpClass = env->GetObjectClass(jni_members.messagePrinter);
   // Retrieve the print method from the messagePrinter
   jni_members.print = env->GetMethodID(jmpClass, "print", "(Ljava/lang/String;)V");
   if (jni_members.print == NULL) {
      env->FatalError("Method MessagePrinter->print(String) not found");
   }

   // Retrieve the class of the KeyData object
   jclass jkdClass = env->GetObjectClass(jkeyData);
   // Retrieve the getDiscID method
   jni_members.getDiscId = env->GetMethodID(jkdClass, "getDiscId", "()[B");
   if (jni_members.getDiscId == NULL) {
      env->FatalError("Method KeyData->getDiscId() not found");
   }
   // Retrieve the setMek method
   jni_members.setMek = env->GetMethodID(jkdClass, "setMek", "([BI)V");
   if (jni_members.setMek == NULL) {
      env->FatalError("Method KeyData->setMek(byte[], int) not found");
   }
   // Retrieve the setVid method
   jni_members.setVid = env->GetMethodID(jkdClass, "setVid", "([BI)V");
   if (jni_members.setVid == NULL) {
      env->FatalError("Method KeyData->setVid(byte[], int) not found");
   }
   // Retrieve the setVuk method
   jni_members.setVuk = env->GetMethodID(jkdClass, "setVuk", "([BI)V");
   if (jni_members.setVuk == NULL) {
      env->FatalError("Method KeyData->setVuk(byte[], int) not found");
   }
   // Retrieve the setTuk method
   jni_members.setTuk = env->GetMethodID(jkdClass, "setTuk", "(I[BI)V");
   if (jni_members.setTuk == NULL) {
      env->FatalError("Method KeyData->setTuk(int, byte[], int) not found");
   }

   // Make a native string from the java string
   const char* path = env->GetStringUTFChars(jpath, NULL);
   if (path == NULL) {
      env->FatalError("Could not access path string from native side");
   }

   // *****************************************************
   // *** Prepare native side to be run in library mode ***
   // *****************************************************
   // Set the global jni struct pointer so that the main function can use the jni stuff
   globalJni = &jni_members;

   // This would be the C++ way to redirect cout, too bad this doesn't affect printf :(
   //class jnistreambuf : public std::stringbuf {
   //public:
   //   jnistreambuf() : std::stringbuf(std::ios_base::out) {}
   //protected:
   //   virtual int sync() {
   //      globalJni->env->CallVoidMethod(globalJni->messagePrinter, globalJni->print, globalJni->env->NewStringUTF(str().c_str()));
   //      return 0;
   //   }
   //};
   //
   //jnistreambuf jnibuf;
   //std::streambuf *coutBufBackup = std::cout.rdbuf();
   //std::cout.rdbuf(&jnibuf);

   // Store the current target of the printf-redirector
   int (*rdprintfBackup)(const char *fmt, ...) = rdprintf;
   // Set function pointer for console output redirection to the java-side function
   rdprintf = &printfj;

   // *******************************************
   // *** Execute native side in library mode ***
   // *******************************************
   // Create the cmdline parameters
   const char* arguments[] = { "moo", "-v", path };
   // Call the main function
   // After this method returns there may be a pending java side exception (in case of an error)!
   // TODO: If for some reason main modifies the arguments, this can lead to a problem
   main(3, const_cast<char**>(arguments));
   // Flush stdout so that any pending messages are printed out
   //TODO: There should be no output to stdout any more, but maybe some other classes than this one do output there
   fflush(stdout);

   // ********************************
   // *** Restore native side      ***
   // *** TODO: Is this necessary? ***
   // ********************************
   // Reset the globalJni struct pointer
   globalJni = NULL;

   // reset the printf-redirector with the backup
   rdprintf = rdprintfBackup;
   //std::cout.rdbuf(coutBufBackup);

   return;
}
#endif
// *** JNI MOD end ***
