/**
 * MAC.h
 *
 * A simple abstraction for the basic functionality of a message authentication code (MAC) engine.
 *
 * @author Paulo S. L. M. Barreto
 *
 * This software is hereby placed in the public domain.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef __MESSAGEAUTHENTICATIONCODE_H
#define __MESSAGEAUTHENTICATIONCODE_H

#ifndef USUAL_TYPES
#define USUAL_TYPES
typedef unsigned char   byte;
/* *** KenD00 start: conflicts under linux *** */
#if defined(WINDOWS) || defined(_WIN32) || defined(__APPLE__)
   #if defined(__LP64__)
      typedef unsigned int   uint;    /* assuming sizeof(uint) == 4 */
   #else
      typedef unsigned long   uint;   /* assuming sizeof(uint) == 4 */
   #endif
#endif
/* *** KenD00 end *** */
#endif /* USUAL_TYPES */

class MAC {

public:
    
    /* *** KenD00: To avoid warnings *** */
    virtual ~MAC() {};

    virtual void init() = 0;

    /**
     * Update the MAC tag computation with a message (AAD or ciphertext) chunk.
     * @param   M   AAD or ciphertext chunk
     * @param   m   its length in bytes
     * @param   aad whether the message chunk is part of the AAD (or else the ciphertext)
     */
    virtual void update(const byte *M, uint m, bool aad) = 0;
    virtual void update(const byte *M) = 0;

    /**
     * Complete a phase (AAD or ciphertext) of the MAC computation.
     * @param   aad whether the message chunk is part of the AAD (or else the ciphertext)
     */
    virtual void finish(bool aad) = 0;

    /**
     * Get the complete MAC tag of the whole message provided.
     * @param   tag the MAC tag
     * @param   t   its length in bytes
     */
    virtual void getTag(byte *tag, uint t) = 0;

};

#endif /* __MESSAGEAUTHENTICATIONCODE_H */
